# -*- coding: utf-8 -*-
"""
Created on Mon Jun  3 03:55:04 2019

@author: iason
"""
import numpy as np
import traceAnalysisCode as analysis
import pandas as pd
import os

#main_path = './traces'
#os.chdir(os.path.dirname(os.path.abspath(__file__)))

main_path = './iasonas/cas9_simulations/DNA11-20_30exposure'
exp = analysis.Experiment(main_path, 30)


for file in exp.files:
    file.autoThreshold('red', threshold=80)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
sys.path.insert(0, os.path.abspath('..'))

# -- Project information -----------------------------------------------------

project = 'papylio'
copyright = '2025 - Chirlmin Joo lab'
# author = 'Ivo Severins, Sung Hyun Kim, Carolien Bastiaanssen, Iason Katechis, Margreet Docter, Roy Simons, Pim America, '
html_logo = '_static/logo.png'

# from git import Repo
from pathlib2 import Path
# raise ValueError(Path(__file__).parent.parent)
# repo = Repo(Path(__file__).parent.parent)

# sha = repo.head.object.hexsha

# The full version, including alpha/beta/rc tags
# release = f'develop-{sha[0:7]}'
import papylio
release = papylio.__version__

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
# extensions = ['sphinx.ext.napoleon', 'sphinx.ext.autodoc', 'sphinx.ext.autosummary']
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    # "sphinx.ext.coverage",
    # "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    # "sphinx.ext.mathjax",
    # "sphinx.ext.todo",
    # "sphinx.ext.autosectionlabel",
    # "sphinx.ext.githubpages",
    "nbsphinx",
    # "IPython.sphinxext.ipython_directive",
    # "IPython.sphinxext.ipython_console_highlighting",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# autodoc_member_order = 'groupwise'
autosummary_generate = True
autoclass_content = "class"
autodoc_default_options = {
    "members": True,
    "undoc-members": False,
    "inherited-members": False,
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_book_theme' # 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# -- Napoleon settings for docstring processing -------------------------------
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_special_with_doc = False
napoleon_use_param = False
napoleon_use_rtype = False
napoleon_preprocess_types = True
# napoleon_type_aliases = {
#     "scalar": ":term:`scalar`",
#     "sequence": ":term:`sequence`",
#     "callable": ":py:func:`callable`",
#     "file-like": ":term:`file-like <file-like object>`",
#     "array-like": ":term:`array-like <array_like>`",
#     "Path": "~~pathlib.Path",
# }


import pytest
import tifffile
import numpy as np
from papylio.movie.movie import Movie


@pytest.fixture
def movie(shared_datadir):
    movie = Movie(shared_datadir / 'BN_TIRF' / 'TIRF 561 0001.tif')
    movie.rot90 = 1
    return movie


@pytest.fixture
def experiment(shared_datadir):
    from papylio import Experiment
    return Experiment(shared_datadir / 'BN_TIRF')


def test_movie_name(movie):
    assert movie.name == 'TIRF 561 0001'


def test_make_projection_image(movie, shared_datadir):
    image_from_method = movie.make_projection_image(projection_type='average', frame_range=(0, 20), illumination=None, write=True,
                                                    return_image=True, flatten_channels=True)
    assert (shared_datadir / 'BN_TIRF' / 'TIRF 561 0001_ave_f0-20_i0.tif').is_file()
    image_from_file = tifffile.imread(shared_datadir / 'BN_TIRF' / 'TIRF 561 0001_ave_f0-20_i0.tif')
    assert (image_from_file == image_from_method).all()
    image_from_original_file = tifffile.imread(shared_datadir / 'BN_TIRF_output_test_movie' / 'TIRF 561 0001_ave_f0-20_i0.tif')
    assert (image_from_original_file == image_from_method).all()
    raw_images = tifffile.imread(shared_datadir / 'BN_TIRF' / 'TIRF 561 0001.tif', key=range(0, 20))
    raw_images = np.rot90(raw_images, axes=(1,2))
    assert ((image_from_file - raw_images.mean(axis=0)) < 1e-4).all()  # Not sure 1e-4 is accurate enough


def test_make_projection_image_frame_range_out_of_bounds(movie, shared_datadir):
    movie.make_projection_image(projection_type='average', frame_range=(390, 410), illumination=None, write=True,
                                return_image=True, flatten_channels=True)

def test_make_projection_images(movie, shared_datadir):
    movie.make_projection_images(projection_type='average', frame_range=(0,20))


def test_determine_background_correction(experiment, shared_datadir):
    movie = experiment.files[1].movie

    movie.determine_general_background_correction()
    movie.determine_temporal_background_correction()
    movie.determine_spatial_background_correction(size=15)
    assert (shared_datadir / 'BN_TIRF' / 'TIRF 561 0001_corrections.nc').is_file()

    import xarray as xr
    new_dataset = xr.load_dataset(shared_datadir / 'BN_TIRF' / 'TIRF 561 0001_corrections.nc')
    new_dataset = new_dataset[['general_background_correction', 'temporal_background_correction', 'spatial_background_correction']]
    test_dataset = xr.load_dataset(shared_datadir / 'BN_TIRF_output_test_movie' / 'TIRF 561 0001_corrections.nc')
    assert new_dataset.identical(test_dataset)

    movie.make_projection_image()

def test_corrections(experiment):
    movie = experiment.files[1].movie
    movie.make_projection_image()


import numpy as np
import matplotlib.pyplot as plt

import itertools
from scipy.spatial import cKDTree
import random
import matchpoint as mp
from skimage.transform import AffineTransform
import time


class GeometricHashTable:
    def __init__(self, destinations, source_vertices=None, initial_source_transformation=AffineTransform(),
                 number_of_source_bases=20, number_of_destination_bases='all',
                 tuple_size=4, maximum_distance_source=None, maximum_distance_destination=None):
        self.initial_source_transformation = initial_source_transformation

        self.tuple_size = tuple_size
        self.maximum_distance_source = maximum_distance_source
        self.maximum_distance_destination = maximum_distance_destination

        # self.mode = mode
        # if mode == 'translation': self.hashTableRange = [-10000, 10000]
        # else: self.hashTableRange = [-1,1]
        # self.nBins = nBins

        #
        # self.number_of_source_bases = number_of_source_bases
        # self.number_of_destination_bases = number_of_destination_bases

        # self._hashTable = None
        # self._matches = None
        #
        # self.rotationRange = rotationRange
        # self.magnificationRange = magnificationRange
        #

        self.destinations = destinations
        self.destination_KDTrees = [cKDTree(destination) for destination in destinations]
        self.source_vertices = source_vertices

        self.number_of_hashtable_entries_per_destination = []
        self.number_of_hashtable_entries_per_basis = []
        self.number_of_bases_per_destination = []

        self.create_hashtable()

    def create_hashtable(self):
        destination_hash_data = geometric_hash(self.destinations, self.maximum_distance_destination, self.tuple_size)
        self.destination_KDTrees, self.destination_tuple_sets, self.destination_hash_table_KDTree, \
        self.destination_transformation_matrices = destination_hash_data


    def query(self, source, distance=15, alpha=0.9, sigma=10, K_threshold=10e9, hash_table_distance_threshold=0.01,
              magnification_range=None, rotation_range=None):

        return find_match_after_hashing(source, self.maximum_distance_source, self.tuple_size, source_vertices,
                                        self.destination_KDTrees, self.destination_tuple_sets,
                                        self.destination_hash_table_KDTree,
                                        hash_table_distance_threshold, alpha, sigma, K_threshold,
                                        magnification_range, rotation_range)

    def query_tuple_transformations(self, sources, hash_table_distance_threshold=0.01, parameters=['rotation', 'scale'],
                                    bins=200):
        # np.vstack(sources)

        source_hash_data = geometric_hash(sources, self.maximum_distance_source, self.tuple_size)
        _, _, source_hash_table_KDTree, source_transformation_matrices = source_hash_data

        return compare_tuple_transformations(source_hash_table_KDTree, source_transformation_matrices,
                                              self.destination_hash_table_KDTree, self.destination_transformation_matrices,
                                              hash_table_distance_threshold, parameters, bins)

    def test(self, sources, hash_table_distance_threshold=0.01, bins=50):
        # np.vstack(sources)

        source_hash_data = geometric_hash(sources, self.maximum_distance_source, self.tuple_size)
        source_KDTrees, source_tuple_sets, source_hash_table_KDTree, source_transformation_matrices = source_hash_data

        tuple_matches = source_hash_table_KDTree.query_ball_tree(self.destination_hash_table_KDTree,
                                                                 hash_table_distance_threshold)

        destination_basis_points = np.array([self.destination_tuple_sets[0][i][0:2] for tuple_match in tuple_matches for i in tuple_match])
        basis_centers = self.destinations[0][destination_basis_points].mean(axis=1)
        plt.scatter(basis_centers[:, 0], basis_centers[:, 1], marker='.')

        plt.figure()
        h, edges = np.histogramdd(basis_centers, bins=bins)
        bin_centers = [(e[:-1] + e[1:]) / 2 for e in edges]
        plt.imshow(h)

        destination_basis_points = np.array([[i]+list(self.destination_tuple_sets[0][j][0:2]) for i, tuple_match in enumerate(tuple_matches) for j in tuple_match])




        linked_tuple_indices = np.array([[i, j] for i, tuple_match in enumerate(tuple_matches) for j in tuple_match])

        source_tuple_centers = source_KDTrees[0].data[np.array(source_tuple_sets[0])].mean(axis=1)
        destination_tuple_centers = self.destination_KDTrees[0].data[np.array(self.destination_tuple_sets[0])].mean(axis=1)

        linked_tuple_centers = np.hstack([source_tuple_centers[linked_tuple_indices[:,0]],destination_tuple_centers[linked_tuple_indices[:,1]]])

        ttt = cKDTree(linked_tuple_centers)
        a = np.unique([j for i in ttt.query_pairs(r=10) for j in i])

        tt1 = cKDTree(linked_tuple_centers[:,:2])
        tt2 = cKDTree(linked_tuple_centers[:,2:])
        a1 = tt1.query_pairs(1)
        a2 = tt2.query_pairs(1)
        ta = np.array(list(a1.intersection(a2)))

        plt.figure()
        c_before = destination_tuple_centers[linked_tuple_indices[:,1]]
        c_after = destination_tuple_centers[linked_tuple_indices[a,1]]
        plt.scatter(self.destination_KDTrees[0].data[:,0], self.destination_KDTrees[0].data[:,1])
        plt.scatter(c_before[:,0], c_before[:,1], c='g')
        plt.scatter(c_after[:,0], c_after[:,1], c='r')



        transformation_matrices = []
        for source_index, destination_indices in linked_tuple_indices[a]:
            source_transformation_matrix = source_transformation_matrices[source_index]
            # source_transformation_matrix = np.linalg.inv(source_transformation_matrix)
            for destination_index in destination_indices:
                destination_transformation_matrix = self.destination_transformation_matrices[destination_index]
                destination_transformation_matrix_inverse = np.linalg.inv(destination_transformation_matrix)
                transformation_matrices.append(destination_transformation_matrix_inverse @ source_transformation_matrix)
        transformation_matrices = np.stack(transformation_matrices)

        # plt.figure()
        # plt.hist(transformation_matrices[:, 0, 2], 100)

        transformations = [AffineTransform(transformation_matrix) for transformation_matrix in transformation_matrices]

        parameter_values = [np.array([getattr(t, parameter) for t in transformations]).T for parameter in parameters]


def mapToPoint(pointSet, startPoints, endPoints, returnTransformationMatrix=False, tr=None, di=None, ro=None):
    startPoints = np.atleast_2d(startPoints)
    endPoints = np.atleast_2d(endPoints)
    if len(startPoints) == 1 & len(endPoints) == 1:
        tr = True;
        ro = False;
        di = False

    elif len(startPoints) == 2 & len(endPoints) == 2:
        if tr is None: tr = True
        if di is None: di = True
        if ro is None: ro = True

    transformationMatrix = np.identity(3)

    if tr:
        translationMatrix = mp.coordinate_transformations.translate(endPoints[0] - startPoints[0])
        transformationMatrix = translationMatrix @ transformationMatrix

    if di or ro:
        diffs = np.array([startPoints[0] - startPoints[1], endPoints[0] - endPoints[1]])
        diffLengths = np.linalg.norm(diffs, axis=1, keepdims=True)
        unitDiffs = diffs / diffLengths

        if di:
            dilationMatrix = mp.coordinate_transformations.magnify(diffLengths[1] / diffLengths[0], endPoints[0])
            transformationMatrix = dilationMatrix @ transformationMatrix

        if ro:
            angle = -np.arctan2(np.linalg.det(unitDiffs), np.dot(unitDiffs[0], unitDiffs[1]))
            # angle = np.arccos(np.dot(diffs[0]/endLength,diffs[1]/startLength))
            rotationMatrix = mp.coordinate_transformations.rotate(angle, endPoints[0])
            transformationMatrix = rotationMatrix @ transformationMatrix

    pointSet = np.append(pointSet, np.ones((pointSet.shape[0], 1)), axis=1)
    transformedPointSet = (transformationMatrix @ pointSet.T)[0:2, :].T

    if returnTransformationMatrix:
        return transformedPointSet, transformationMatrix
    else:
        return transformedPointSet


def compare_tuple_transformations(source_hash_table_KDTree, source_transformation_matrices, destination_hash_table_KDTree,
                                  destination_transformation_matrices, hash_table_distance_threshold=0.01,
                                  parameters=['rotation', 'scale'], bins=200):
    tuple_matches = source_hash_table_KDTree.query_ball_tree(destination_hash_table_KDTree, hash_table_distance_threshold)

    # TODO: make this matrix multiplication
    transformation_matrices = []
    for source_index, destination_indices in enumerate(tuple_matches):
        source_transformation_matrix = source_transformation_matrices[source_index]
        # source_transformation_matrix = np.linalg.inv(source_transformation_matrix)
        for destination_index in destination_indices:
            destination_transformation_matrix = destination_transformation_matrices[destination_index]
            destination_transformation_matrix_inverse = np.linalg.inv(destination_transformation_matrix)
            transformation_matrices.append(destination_transformation_matrix_inverse @ source_transformation_matrix)
    transformation_matrices = np.stack(transformation_matrices)

    # plt.figure()
    # plt.hist(transformation_matrices[:, 0, 2], 100)

    transformations = [AffineTransform(transformation_matrix) for transformation_matrix in transformation_matrices]


    parameter_values = [np.array([getattr(t, parameter) for t in transformations]).T for parameter in parameters]

    sample = np.vstack(list(parameter_values)).T
    # sample = transformation_matrices[:, :2, :].reshape(-1, 6)

    h, edges = np.histogramdd(sample, bins=bins)

    bin_centers = [(e[:-1]+e[1:])/2 for e in edges]
    # found_transformation = np.array([bc[h_index] for bc, h_index in zip(bin_centers, np.where(h==h.max()))]).reshape(2,3)
    # t = AffineTransform(np.vstack([found_transformation, [0, 0, 1]]))
    hist_max_index = np.where(h == h.max())
    if len(hist_max_index[0]) > 1:
        raise RuntimeError('No optimal transformation found')
    found_values = [bc[h_index][0] for bc, h_index in zip(bin_centers, hist_max_index)]

    parameter_dict = {parameter: found_values.pop(0) if parameter == 'rotation' else [found_values.pop(0), found_values.pop(0)]
                      for parameter in parameters}

    # found_transformation = AffineTransform(**parameter_dict)

    # Hc = H.copy()
    # Hc[i]=0
    # print(np.max(H)/np.max(Hc)>2)

    # x, y = bin_centers_x, bin_centers_y
    # X, Y = np.meshgrid(x, y)
    # xdata = np.vstack((X.ravel(), Y.ravel()))

    # def twoD_gaussian(M, offset, amplitude, x0, y0, sigma_x, sigma_y):
    #     x, y = M
    #     return offset + amplitude * np.exp(- ((x - x0) / (2 * sigma_x)) ** 2 - ((y - y0) / (2 * sigma_y)) ** 2)
    #
    # from scipy.optimize import curve_fit
    # p0 = [20,20,0,0,1,1]
    # popt, pcov = curve_fit(twoD_gaussian, xdata, H.ravel())#, p0) #input: function, xdata, ydata,p0
    #
    # coeff, var_matrix = curve_fit(gauss.mult_gaussFun_Fit, (bin_centers_x, bin_centers_y), H, p0=p0)

    # plt.figure()
    # plt.scatter(ms, rs)
    #
    # plt.figure()
    # plt.hist(ms, 100)
    #
    # plt.figure()
    # plt.hist(rs, 100)

    return parameter_dict





#
# t0 = time.time()
# from matchpoint.geometricHashing import pointHash, findMatch
# ht = pointHash(destination, bases='all', magnificationRange=[0,10], rotationRange=[-np.pi,np.pi])
# matched_bases = findMatch(source, ht, bases='all', magnificationRange=[0,1], rotationRange=[-np.pi,np.pi])
# source_coordinate_tuple = source[matched_bases['testBasis']]
# destination_coordinate_tuple = destination[matched_bases['hashTableBasis']]
# t1 = time.time()
# plt.close('all')


# centers = np.array([(destination[pair[0]]+destination[pair[1]])/2 for pair in pairs])
#
# centers_KDTree = KDTree(centers)
#
# # Points within the circle containing the two original points in the pair
# indices_in_between_pairs = centers_KDTree.query_ball_tree(destination_KDTree, r=distance/2*0.8)


# def generate_point_tuples(point_set_KDTree, maximum_distance, tuple_size):
#     pairs = list(point_set_KDTree.query_pairs(maximum_distance))
#
#     point_tuples = []
#
#     for pair in pairs:
#         pair_coordinates = point_set_KDTree.data[list(pair)]
#         center = (pair_coordinates[0] + pair_coordinates[1]) / 2
#         distance = np.linalg.norm(pair_coordinates[0] - pair_coordinates[1])
#         internal_points = point_set_KDTree.query_ball_point(center, distance / 2 * 0.99)
#         if len(internal_points) >= (tuple_size - 2):
#             random.shuffle(internal_points)
#             internal_points = internal_points[0:(tuple_size - 2)]
#             #point_tuples.append(pair + tuple(internal_points))
#             yield pair + tuple(internal_points)
#
#     #return point_tuples

def generate_point_tuples(point_set_KDTree, maximum_distance, tuple_size):
    pairs = list(point_set_KDTree.query_pairs(maximum_distance))

    point_tuples = []

    for pair in pairs:
        pair_coordinates = point_set_KDTree.data[list(pair)]
        center = (pair_coordinates[0] + pair_coordinates[1]) / 2
        distance = np.linalg.norm(pair_coordinates[0] - pair_coordinates[1])
        internal_points = point_set_KDTree.query_ball_point(center, distance / 2 * 0.99)
        for internal_point_combination in itertools.combinations(internal_points, tuple_size-2):
            point_tuples.append(pair + tuple(internal_point_combination))
            # yield pair + tuple(internal_point_combination)

    return point_tuples

def geometric_hash_table(point_set_KDTree, point_tuples):
    pt = np.array(point_tuples)
    d = point_set_KDTree.data[pt]
    d0 = np.array([[0,0],[1,1]])


    T = np.repeat(np.expand_dims(np.diag([1.,1.,1.],k=0), axis=0), len(d), axis=0)
    T[:, :2, 2] = d0[0] - d[:, 0, :]


    diffs_d = d[:,1,:] - d[:,0,:]
    diffs_d0 = d0[1] - d0[0]
    m_d = np.linalg.norm(diffs_d, axis=1)
    m_d0 = np.linalg.norm(diffs_d0)

    unit_diffs_d = np.divide(diffs_d, m_d[:, None])
    unit_diffs_d0 = diffs_d0 / m_d0

    D = np.repeat(np.expand_dims(np.diag([1., 1., 1.], k=0), axis=0), len(d), axis=0)
    D[:,0,0] = D[:,1,1] = 1 / m_d * m_d0

    dot_product = np.dot(unit_diffs_d, unit_diffs_d0) # for mapping to (0,0),(1,1) could be replaced by sum to increase speed
    cross_product = np.cross(unit_diffs_d, unit_diffs_d0) # for mapping to (0,0),(1,1) could be replaced by subtraction to increase speed
    R = np.repeat(np.expand_dims(np.diag([1., 1., 1.], k=0), axis=0), len(d), axis=0)
    R[:,0,0] = R[:,1,1] = dot_product
    R[:,1,0] = cross_product
    R[:,0,1] = -cross_product

    transformation_matrices = R@D@T

    # transformation_matrices @ np.dstack([d, np.ones((d.shape[0], d.shape[1], 1))]).swapaxes(1,2)
    hash_code = transformation_matrices[:,:2,:] @ np.dstack([d[:,len(d0):,:], np.ones((d.shape[0], d.shape[1]-len(d0), 1))]).swapaxes(1, 2)

    # Break similarity of pair
    break_similarity = hash_code.swapaxes(1,2)[:, :, 0].sum(axis=1) > ((pt.shape[1] - d0.shape[0]) / 2)
    # this is not yet suited for using more than two basis points
    switch_basis_points_matrix = np.array([[-1,0,1],[0,-1,1],[0,0,1]])
    transformation_matrices[break_similarity] = switch_basis_points_matrix[None,:,:]@transformation_matrices[break_similarity]
    hash_code[break_similarity] = transformation_matrices[break_similarity, :2, :] @ np.dstack(
        [d[break_similarity, len(d0):, :], np.ones((d[break_similarity].shape[0], d.shape[1] - len(d0), 1))]).swapaxes(1, 2)

    hash_code = hash_code.swapaxes(1, 2)

    # Break similarity of internal points
    s = hash_code[:, :, 0].argsort()
    hash_code = np.take_along_axis(hash_code, s[:, :, None], axis=1)

    return hash_code.reshape(len(hash_code), -1), transformation_matrices

    # x = np.hstack([d[0][0:2], [[1],[1]]])
    # y = np.hstack([d0, [[1],[1]]])
    #
    # x = np.vstack([x, x[1]-x[0]])
    # y = np.vstack([y, y[1]-y[0]])
    #
    # A = np.linalg.lstsq(x.T, y.T, rcond=None)[0]
    #
    # scale = np.sqrt(x * x).sum(axis=0)
    # coeff, r, rank, s = np.linalg.lstsq(x / scale, y[:, 0:2], rcond=None)
    # A = (coeff.T / scale).T

### Old version
# def geometric_hash_table(point_set_KDTree, point_tuples, tuple_size):
#     #hash_table = []
#     for point_tuple in point_tuples:
#
#         pair = point_tuple[:2]
#         internal_points = point_tuple[2:]
#         pair_coordinates = point_set_KDTree.data[list(pair)]
#         internal_coordinates = point_set_KDTree.data[list(internal_points)]
#         # Sort internal points based on x coordinate
#         # internal_point_order = np.argsort(internal_coordinates[:, 0])
#         # internal_points = [internal_points[i] for i in internal_point_order]
#         # internal_coordinates = internal_coordinates[internal_point_order]
#
#         end_points = np.array([[0, 0], [1, 1]])
#         hash_coordinates = mapToPoint(internal_coordinates, pair_coordinates, end_points)
#         # Break similarity of pair
#         if np.sum(hash_coordinates[:, 0]) > ((tuple_size - 2) / 2):
#             pair = pair[::-1]
#             pair_coordinates = pair_coordinates[::-1]
#             hash_coordinates = mapToPoint(internal_coordinates, pair_coordinates, end_points)
#
#         # Break similarity of internal points
#         internal_point_order = np.argsort(hash_coordinates[:, 0])
#         internal_points = [internal_points[i] for i in internal_point_order]
#         internal_coordinates = internal_coordinates[internal_point_order]
#         hash_coordinates = hash_coordinates[internal_point_order]
#
#         point_tuple = pair + tuple(internal_points)
#         hash_code = hash_coordinates.flatten()
#         # hash_table.append(hash_code)
#
#         yield point_tuple, hash_code
#
#         # plot_tuple(np.vstack([pair_coordinates, internal_coordinates])
#         # plot_tuple(np.vstack([end_points, hash_coordinates]))

def geometric_hash(point_sets, maximum_distance=100, tuple_size=4):
    # TODO: Add minimum_distance and implement
    # TODO: Make invariant to mirroring
    # TODO: Make usable with multiple point-sets in a single hash table
    # TODO: Implement names of point_sets, possibly through a dictionary and adding a attribute to each KDtree
    start_time = time.time()

    if type(point_sets) is not list:
        point_sets = [point_sets]

    point_set_KDTrees = [cKDTree(point_set) for point_set in point_sets]

    hash_tables = []
    point_tuple_sets = []
    for point_set_KDTree in point_set_KDTrees:
        point_tuples = generate_point_tuples(point_set_KDTree, maximum_distance, tuple_size)
        hash_table, transformation_matrices = geometric_hash_table(point_set_KDTree, point_tuples)

        hash_tables.append(hash_table)
        point_tuple_sets.append(point_tuples)

    hash_table_KDTree = cKDTree(np.vstack(hash_tables))

    print("--- %s seconds ---" % (time.time() - start_time))

    return point_set_KDTrees, point_tuple_sets, hash_table_KDTree, transformation_matrices

### Old version
# def geometric_hash(point_sets, maximum_distance=100, tuple_size=4):
#     # TODO: Add minimum_distance and implement
#     # TODO: Make invariant to mirroring
#     # TODO: Make usable with multiple point-sets in a single hash table
#     # TODO: Implement names of point_sets, possibly through a dictionary and adding a attribute to each KDtree
#     start_time = time.time()
#
#     if type(point_sets) is not list:
#         point_sets = [point_sets]
#
#     point_set_KDTrees = [cKDTree(point_set) for point_set in point_sets]
#
#     hash_table = []
#     point_tuple_sets = []
#     for point_set_KDTree in point_set_KDTrees:
#         point_tuple_generator = generate_point_tuples(point_set_KDTree, maximum_distance, tuple_size)
#
#         point_tuples = []
#         for point_tuple, hash_code in geometric_hash_table(point_set_KDTree, point_tuple_generator, tuple_size):
#             point_tuples.append(point_tuple)
#             hash_table.append(hash_code)
#
#         point_tuple_sets.append(point_tuples)
#
#     hash_table_KDTree = cKDTree(np.array(hash_table))
#
#     print("--- %s seconds ---" % (time.time() - start_time))
#
#     return point_set_KDTrees, point_tuple_sets, hash_table_KDTree




def find_match_after_hashing(source, maximum_distance_source, tuple_size, source_vertices,
                             destination_KDTrees, destination_tuple_sets, destination_hash_table_KDTree,
                             hash_table_distance_threshold=0.01,
                             alpha=0.1, test_radius=10, K_threshold=10e9,
                             magnification_range=None, rotation_range=None):

    if type(destination_KDTrees) is not list:
        destination_KDTrees = [destination_KDTrees]

    source_KDTree = cKDTree(source)
    source_tuple_generator = generate_point_tuples(source_KDTree, maximum_distance_source, tuple_size)

    hash_table_distances_checked = 0
    tuples_checked = 0
    # for source_tuple_index in np.arange(len(source_tuples)):
    for source_tuple, source_hash_code in geometric_hash_table(source_KDTree, source_tuple_generator, tuple_size):
        # TODO: Get all destination tuple indices within a range
        distance, destination_tuple_index = destination_hash_table_KDTree.query(source_hash_code)

        # We can also put a threshold on the distance here possibly
        #print(distance)
        hash_table_distances_checked += 1
        # if hash_table_distances_checked > 500:
        #     return

        if distance < hash_table_distance_threshold:
            # source_coordinate_tuple = source[list(source_tuples[source_tuple_index])]
            # destination_coordinate_tuple = destination[list(destination_tuples[destination_tuple_index])]

            # source_tuple = source_tuples[source_tuple_index]
            # Find the destination tuple by determining in which destination pointset the destination tuple index is located
            # and by determining what the index is within that destination pointset.
            cumulative_tuples_per_destination = np.cumsum([0]+[len(point_tuples) for point_tuples in destination_tuple_sets])
            destination_index = np.where((cumulative_tuples_per_destination[:-1] <= destination_tuple_index) &
                                         (cumulative_tuples_per_destination[1:] > destination_tuple_index))[0][0]
            tuple_index_in_destination_set = destination_tuple_index - cumulative_tuples_per_destination[destination_index]
            destination_tuple = destination_tuple_sets[destination_index][tuple_index_in_destination_set]

            #Or list(itertools.chain.from_iterable(destination_tuple_sets))[destination_tuple_index]

            tuples_checked += 1

            destination_KDTree = destination_KDTrees[destination_index]
            found_transformation = tuple_match(source, destination_KDTree, source_vertices, source_tuple, destination_tuple,
                                alpha, test_radius, K_threshold, magnification_range, rotation_range)
            if found_transformation:
                match = mp.MatchPoint(source=source, destination=destination_KDTree.data, method='Geometric hashing',
                                 transformation_type='linear', initial_transformation=None)
                match.transformation = found_transformation
                match.destination_index = destination_index

                match.hash_table_distance = distance
                match.hash_table_distances_checked = hash_table_distances_checked
                match.tuples_checked = tuples_checked
                return match

def polygon_area(vertices):
    x = vertices[:,0]
    y = vertices[:,1]
    return 0.5 * np.abs(np.dot(x, np.roll(y, 1)) - np.dot(y, np.roll(x, 1)))

def tuple_match(source, destination_KDTree, source_vertices, source_tuple, destination_tuple,
                alpha=0.1, test_radius=10, K_threshold=10e9, scaling_range=None, rotation_range=None):
    source_coordinate_tuple = source[list(source_tuple)]
    destination_coordinate_tuple = destination_KDTree.data[list(destination_tuple)]

    source_indices_without_tuple = [i for i in range(len(source)) if i not in source_tuple]
    source = source[source_indices_without_tuple]

    source_transformed, transformation_matrix = mapToPoint(source, source_coordinate_tuple[:2], destination_coordinate_tuple[:2], returnTransformationMatrix=True)
    # mp.icp.scatter_coordinates([source_transformed])

    found_transformation = AffineTransform(transformation_matrix)

    if rotation_range:
        rotation = found_transformation.rotation/(2*np.pi)*360
        if not (rotation_range[0] < rotation < rotation_range[1]):
            return
    if scaling_range:
        scaling = np.mean(found_transformation.scale)
        if not (scaling_range[0] < scaling < scaling_range[1]):
            return

    # if not (-1 < rot < 1) & (3.3 < sca < 3.4):
    #     return
    # if (-1 < rot < 1) & (3.3 < sca < 3.4):
    #     print('found')
    # mp.icp.scatter_coordinates([destination_KDTree.data, source_transformed])

    source_vertices_transformed = found_transformation(source_vertices)
    destination_cropped = mp.point_set.crop_coordinates(destination_KDTree.data, source_vertices_transformed)

    #source_transformed_area = np.linalg.norm(source_vertices_transformed[0] - source_vertices_transformed[1])
    # source_transformed_area = np.abs(np.cross(source_vertices_transformed[1] - source_vertices_transformed[0],
    #                                           source_vertices_transformed[3] - source_vertices_transformed[0]))

    source_transformed_area = polygon_area(source_vertices_transformed)

    # pDB = 1 / source_transformed_area
    # pDB = 1 / len(destination_cropped)
    # pDB = 1
    # pDB = 1 / len(source)
    pDB = 1 / source_transformed_area

    # alpha = 0.1
    # # test_radius = 5
    # test_radius = 10
    K=1
    # K_threshold = 10e9
    # bayes_factor6 = []
    for coordinate in source_transformed:
        # points_within_radius = destination_KDTree.query_ball_point(coordinate, test_radius)
        # pDF = alpha/source_transformed_area + (1-alpha)*len(points_within_radius)/len(destination_cropped)
        # pDF = alpha/len(destination_cropped) + (1-alpha)*len(points_within_radius)/len(destination_cropped)
        # pDF = alpha + (1-alpha)*len(points_within_radius)
        # pDF = alpha/len(source) + (1-alpha)*len(points_within_radius)/len(destination_cropped)
        # pDF = alpha/source_transformed_area + (1-alpha)*len(points_within_radius)/(len(destination_cropped)*np.pi*test_radius**2)

        sigma = test_radius
        # sigma = 10
        distance, index = destination_KDTree.query(coordinate)
        # pDF = alpha/source_transformed_area + (1-alpha)/(2*np.pi*sigma**2)*np.exp(-distance**2/(2*sigma**2))/len(destination_cropped)
        # 2d Gaussian
        pDF = alpha / source_transformed_area + \
              (1 - alpha) / (2 * np.pi * sigma ** 2) * \
              np.exp(-(distance ** 2) / (2 * sigma ** 2)) / len(destination_cropped)

        # print(len(points_within_radius))
        # bayes_factor6.append(pDF/pDB)
        K = K * pDF/pDB
        if K > K_threshold:
            print("Found match")
            return found_transformation

    # print(pDF)
    # print(K)



def find_match(source, destination, source_vertices,
               tuple_size=4, maximum_distance_source=4, maximum_distance_destination=40,
               hash_table_distance_threshold = 0.01,
               alpha = 0.1, test_radius = 10, K_threshold = 10e9
            ):
    # destination_KDTree, destination_tuples, destination_hash_table_KDTree = geometric_hash(destination, 40, 4)
    # source_KDTree, source_tuples, source_hash_table_KDTree = geometric_hash(source, 4, 4)
    # 200 points 200,20
    # 10000 points 10,1
    # return find_match_after_hashing(*geometric_hash(source, 4, 4), source_vertices, *geometric_hash(destination, 40, 4))


    return find_match_after_hashing(source, maximum_distance_source, tuple_size, source_vertices,
                                    *geometric_hash(destination, maximum_distance_destination, tuple_size),
                                    hash_table_distance_threshold,
                                    alpha, test_radius, K_threshold)


if __name__ == '__main__':
    from papylio.plugins.sequencing.point_set_simulation import simulate_mapping_test_point_set

    # Simulate source and destination point sets
    number_of_source_points = 4000
    transformation = AffineTransform(translation=[128, 128], rotation=0 / 360 * 2 * np.pi, scale=[1, 1])
    source_bounds = np.array([[0, 0], [512, 512]])
    source_crop_bounds = np.array([[0, 0], [50, 50]])
    fraction_missing_source = 0
    fraction_missing_destination = 0
    maximum_error_source = 0
    maximum_error_destination = 0
    shuffle = True

    destination, source = simulate_mapping_test_point_set(number_of_source_points, transformation,
                                                          source_bounds, source_crop_bounds,
                                                          fraction_missing_source, fraction_missing_destination,
                                                          maximum_error_source, maximum_error_destination, shuffle)
    destinations = [destination]

    perfect = mp.MatchPoint(source, destination)
    perfect.transformation = AffineTransform(matrix=transformation._inv_matrix)
    perfect.show_mapping_transformation(show_source=True)

    # mp.icp.scatter_coordinates([source, destination])

    source_vertices = np.array([source_crop_bounds[0], source_crop_bounds.T[0],
                                source_crop_bounds[1], np.flip(source_crop_bounds.T[1])])




    # destination_KDTree, destination_tuples, destination_hash_table_KDTree = geometric_hash(destination, 3000, 4)
    # source_KDTree, source_tuples, source_hash_table_KDTree = geometric_hash(source, 100, 4)





    hash_table_distance_threshold = 0.01
    alpha = 0.5
    test_radius = 10
    K_threshold = 10e9

    # match = find_match_after_hashing(source, maximum_distance_source, tuple_size, source_vertices,
    #                                  *destination_hash_data2,
    #                                  hash_table_distance_threshold=0.01,
    #                                  alpha=0.1, test_radius=10, K_threshold=10e9)

    maximum_distance_destination = 20
    maximum_distance_source = 20


    # ht = GeometricHashTable(destinations, source_vertices, initial_source_transformation=AffineTransform(),
    #                  number_of_source_bases=20, number_of_destination_bases='all',
    #                  tuple_size=4, maximum_distance_source=maximum_distance_source, maximum_distance_destination=maximum_distance_destination)
    #
    #
    # # test = ht.query_tuple_transformations(source, hash_table_distance_threshold=0.1, parameters=['rotation'], bins=50)
    # plt.figure()
    # t = ht.test(source, hash_table_distance_threshold=0.01, bins=25)




    import matplotlib.pyplot as plt
    import matplotlib.tri as tri
    import numpy as np

    triang_source = tri.Triangulation(source[:, 0], source[:, 1])

    triang_destination = tri.Triangulation(destination[:, 0], destination[:, 1])





    def angles_from_triangles(pointset, triangles):
        triangle_points = pointset[triangles.triangles]

        side_lengths = np.linalg.norm(triangle_points[:, [0, 1, 2], :] - triangle_points[:, [1, 2, 0], :], axis=2)

        a = side_lengths[:, 0]
        b = side_lengths[:, 1]
        c = side_lengths[:, 2]

        # R = a * b * c / np.sqrt((a + b + c) * (a - b + c) * (a + b - c) * (b + c - a))
        # angles = np.arcsin(side_lengths / (2 * R[:, np.newaxis]))

        A = np.arccos((-a ** 2 + b ** 2 + c ** 2) / (2 * b * c))
        B = np.arccos((a ** 2 - b ** 2 + c ** 2) / (2 * c * a))
        C = np.arccos((a ** 2 + b ** 2 - c ** 2) / (2 * a * b))

        A = (-a ** 2 + b ** 2 + c ** 2) / (2 * b * c)
        B = (a ** 2 - b ** 2 + c ** 2) / (2 * c * a)
        C = (a ** 2 + b ** 2 - c ** 2) / (2 * a * b)

        angles = np.vstack([A, B, C]).T

        return angles

    angles_source = angles_from_triangles(source, triang_source)#/(2*np.pi)*360
    angles_destination = angles_from_triangles(destination, triang_destination)#/(2*np.pi)*360

    angles_source.sort()
    angles_destination.sort()

    hash_table_source = cKDTree(angles_source[:, :2])
    hash_table_destination = cKDTree(angles_destination[:,:2])

    matches = hash_table_source.query_ball_tree(hash_table_destination, 0.01)

    [len(i) for i in test]

    triang_source.triangles[51]

    triangle_index = 0

    source_neighbours = triang_source.neighbors[triangle_index]
    source_neighbours_in_destination = [matches[source_triangle_index] if source_triangle_index >= 0 else [] for source_triangle_index in source_neighbours]
    destination_triangles = matches[triangle_index]
    destination_neighbours = triang_destination.neighbors[destination_triangles]

    t2 = np.array(
        [[dn in source_neighbours_in_destination[i] for dn in destination_neighbours[:, i]] for i in range(3)]).T

    # fig1 = plt.figure()
    fig1.clf()
    plot_triangles(triang_source, highlight=[0], figure=fig1)
    plot_triangles(triang_source, highlight=source_neighbours,
                   highlight_kwargs={'c': 'r'}, figure=fig1)

    # fig2 = plt.figure()
    fig2.clf()
    plot_triangles(triang_destination, highlight=matches[0], figure=fig2)
    plot_triangles(triang_destination, highlight=destination_neighbours.flatten(),
                   highlight_kwargs={'c': 'r'}, figure=fig2)

    plot_triangles(triang_destination, highlight=np.hstack(source_neighbours_in_destination),
                   highlight_kwargs={'c': 'k', 'marker': 'o'}, figure=fig2)




    t2s = []
    t2cs = []
    for triangle_index, _ in enumerate(triang_source.triangles):
        source_neighbours = triang_source.neighbors[triangle_index]
        source_neighbours_in_destination = [matches[i] if i >= 0 else [] for i in source_neighbours]
        destination_neighbours = triang_destination.neighbors[matches[triangle_index]]

        t2 = np.array([[dn in source_neighbours_in_destination[i] for dn in destination_neighbours[:, i]] for i in range(3)]).T

        t2 = t2.sum(axis=1)
        t2u = np.unique(t2, return_counts=True)
        t2c = np.zeros(4).astype(int)
        print(t2u)
        t2c[t2u[0]] = t2u[1]
        print(t2)
        t2cs.append(t2c)
        if t2c[3]==1:
            t2s.append((triangle_index, test[triangle_index][np.where(t2==3)[0][0]]))
    t2s = np.vstack(t2s)
    t2cs = np.array(t2cs)


    tr = np.array(t2s)

    mask = np.ones(len(triang_source.triangles)).astype(bool)
    mask[tr[:,0]] = False

    triang_source.set_mask(mask)

    mask = np.ones(len(triang_destination.triangles)).astype(bool)
    mask[tr[:,1]] = False

    triang_destination.set_mask(mask)




    fig2, ax2 = plt.subplots()
    ax2.set_aspect('equal')
    ax2.triplot(triang_destination, 'o-', lw=1)



    def plot_triangles(triangulation, highlight=[], highlight_kwargs={'c': 'b'}, figure=None):
        if not figure:
            figure = plt.figure()
        ax = figure.gca()
        ax.set_aspect('equal')
        ax.triplot(triangulation, '-', lw=1)

        triangle_indices = highlight
        point_indices = triangulation.triangles[triangle_indices]
        x_mean = triangulation.x[point_indices].mean(axis=1)
        y_mean = triangulation.y[point_indices].mean(axis=1)
        ax.scatter(x_mean, y_mean, **highlight_kwargs)

        return figure


    match.show_mapping_transformation(show_source=True)

    match = find_match_after_hashing(*source_hash_data, source_vertices, *destination_hash_data)

    match = find_match(source, destination, source_vertices)
    if match:
        mp.icp.scatter_coordinates([source, destination, match(source), source_vertices, match(source_vertices)])







    #
    # import papylio as pp
    # exp = pp.Experiment(r'D:\SURFdrive\Promotie\Data\Sung hyun and Frank')
    #
    # import numpy as np
    # tile1101 = np.loadtxt(r'D:\SURFdrive\Promotie\Data\Sung hyun and Frank\1101.loc')
    # tile1102 = np.loadtxt(r'D:\SURFdrive\Promotie\Data\Sung hyun and Frank\1102.loc')
    #
    #
    # tuple_size = 4
    # maximum_distance_destination = 1000
    # tile1101_hash_data = geometric_hash2(tile1101, maximum_distance_destination, tuple_size)
    # tile1102_hash_data = geometric_hash2(tile1102, maximum_distance_destination, tuple_size)
    #
    # destination_hash_data = tile1102_hash_data
    # destination_KDTree, destination_tuples, destination_hash_table_KDTree, destination_transformation_matrices = destination_hash_data
    #
    # Ts = []
    # for file in exp.files:
    #
    #     maximum_distance_source = 200
    #     file_coordinates = file.coordinates[1::2]
    #     file_coordinates[:,0] = -file_coordinates[:,0]
    #     source_hash_data = geometric_hash2(file_coordinates, maximum_distance_source, tuple_size)
    #
    #     source_KDTree, source_tuples, source_hash_table_KDTree, source_transformation_matrices = source_hash_data
    #
    #
    #     htmatch = source_hash_table_KDTree.query_ball_tree(destination_hash_table_KDTree, 0.005)
    #
    #     Tss=[]
    #     for i, htm in enumerate(htmatch):
    #         sT = source_transformation_matrices[i]
    #         sTi = np.linalg.inv(sT)
    #         for dm in htm:
    #             dT = destination_transformation_matrices[dm]
    #             dTi = np.linalg.inv(dT)
    #             Tss.append(dTi@sT)
    #             print('appended')
    #     Ts.append(Tss)
    #
    # for Tss in Ts:
    #     Tss = np.stack(Tss)
    #
    #
    #     ms = np.sqrt(Tss[:, 0, 0] ** 2 + Tss[:, 1, 0]**2)
    #     plt.figure()
    #     h, b, _ = plt.hist(ms, 100)
    #
    # m = np.mean(b[h.argmax():h.argmax() + 2])
    #
    # sel = np.all(np.vstack([ms>2.4, ms<2.5]), axis=0)
    # test = Ts[sel]
    #
    # h, b, _ = plt.hist(np.arccos(test[:,0,0]/m), 100)
    # r = np.mean(b[h.argmax():h.argmax() + 2])
    #
    #
    #
    # trs=[]
    # for Tss in Ts:
    #     for tr in Tss:
    #         trs.append(AffineTransform(tr))
    #
    # ms = np.array([tr.scale[0] for tr in trs])
    # rs = np.array([tr.rotation for tr in trs])
    #
    # plt.scatter(ms, rs)
    #
    # plt.hist(ms, 100)
    # plt.hist(rs, 100)
    #
    # plt.close('all')
    # for file in exp.files:
    #     file_coordinates = file.coordinates[1::2]
    #     file_coordinates[:,0] = -file_coordinates[:,0]
    #     source_vertices = np.array([[256,0],[256,512],[512,512],[512,0]])
    #     match = find_match_after_hashing(file_coordinates, maximum_distance_source, tuple_size, source_vertices,
    #                                      *tile1101_hash_data,
    #                                      hash_table_distance_threshold=0.01,
    #                                      alpha=0.5, test_radius=50, K_threshold=10e9)
    #     if match is not None:
    #         match.show_mapping_transformation()
















    # Make random source and destination dataset
    # np.random.seed(42)
    # destination = np.random.rand(1000,2)*1000
    # destination2 = np.random.rand(1000, 2) * 1000
    # source_vertices_in_destination = np.array([[300, 300], [450, 300], [450, 600], [300,600]])
    #
    #
    # transformation = AffineTransform(scale=(0.1, 0.1), rotation=np.pi, shear=None, translation=(-100,350))
    # source = transformation(mp.point_set.crop_coordinates(destination, source_vertices_in_destination))
    # source_vertices = transformation(source_vertices_in_destination)
    # plt.figure()
    # plt.scatter(destination[:,0],destination[:,1])
    # plt.scatter(source[:,0],source[:,1])
    # mp.icp.scatter_coordinates([source,destination,mp.point_set.crop_coordinates(destination, source_vertices_in_destination)])
    #
    # # destination_KDTree, destination_tuples, destination_hash_table_KDTree = geometric_hash(destination, 40, 4)
    # # source_KDTree, source_tuples, source_hash_table_KDTree = geometric_hash(source, 4, 4)
    # #
    # match = find_match(source, [destination2, destination], source_vertices,
    #                    tuple_size=4, maximum_distance_source=4, maximum_distance_destination=40,
    #                    hash_table_distance_threshold = 0.01,
    #                    alpha = 0.1, test_radius = 10, K_threshold = 10e9)
    # plt.figure()
    # mp.icp.scatter_coordinates([source, destination, match(source), source_vertices,source_vertices_in_destination])



    # file_coordinates = np.loadtxt(
    #     r'C:\Users\Ivo Severins\Desktop\seqdemo\20190924 - Single-molecule setup (TIR-I)\16L\spool_6.pks')[:, 1:3]
    # tile_coordinates = np.loadtxt(r'C:\Users\Ivo Severins\Desktop\seqdemo\20190926 - Sequencer (MiSeq)\2102.loc')
    #
    # source = file_coordinates
    # destination = tile_coordinates
    # source_vertices = np.array([[1024,0],[2048,0],[2048,2048],[1024,2048]])
    #
    # # source = file_coordinates[[1,5,9,7]]
    # # destination = tile_coordinates[[73,82,85,84]]
    #
    # source[:,0] = -source[:,0]
    # source_vertices[:,0] = -source_vertices[:,0]
    #
    # # destination_KDTree, destination_tuples, destination_hash_table_KDTree = geometric_hash(destination, 3000, 4)
    # # source_KDTree, source_tuples, source_hash_table_KDTree = geometric_hash(source, 100, 4)
    #
    # tuple_size = 4
    # maximum_distance_source = 1000
    # maximum_distance_destination = 3000
    #
    # source_hash_data = geometric_hash(source, maximum_distance_source, tuple_size)
    # destination_hash_data = geometric_hash(destination, maximum_distance_destination, tuple_size)
    #
    # hash_table_distance_threshold = 0.01
    # alpha = 0.1
    # test_radius = 10
    # K_threshold = 10e9
    #
    # match = find_match_after_hashing(source, maximum_distance_source, tuple_size, source_vertices,
    #                                  *destination_hash_data,
    #                                  hash_table_distance_threshold=0.01,
    #                                  alpha=0.1, test_radius=10, K_threshold=10e9)
    #
    #
    # source_KDTree, source_tuples, source_hash_table_KDTree = source_hash_data
    # destination_KDTree, destination_tuples, destination_hash_table_KDTree = destination_hash_data
    #
    # #match = find_match_after_hashing(*source_hash_data, source_vertices, *destination_hash_data)
    #
    # #match = find_match(source, destination, source_vertices)
    # if match:
    #     mp.icp.scatter_coordinates([source, destination, match(source), source_vertices, match(source_vertices)])

    #
    #
    # def scatter_with_number(coordinates):
    #     # plot the chart
    #     plt.scatter(coordinates[:,0], coordinates[:,1])
    #
    #     # zip joins x and y coordinates in pairs
    #     for i, coordinate in enumerate(coordinates):
    #         label = str(i)
    #
    #         # this method is called for each point
    #         plt.annotate(label,  # this is the text
    #                      (coordinate[0], coordinate[1]),  # this is the point to label
    #                      textcoords="offset points",  # how to position the text
    #                      xytext=(0, 10),  # distance from text to points (x,y)
    #                      ha='center')
    #
    # scatter_with_number(source)
    # scatter_with_number(destination)
    #
    # for i, t in enumerate(source_tuples):
    #     if ((1 in t) & (5 in t) & (9 in t) & (7 in t)):
    #         print(i)
    #
    # for i, t in enumerate(destination_tuples):
    #     if ((73 in t) & (82 in t) & (85 in t) & (84 in t)):
    #         print(i)
    #
    # tuple_match(source, destination_KDTree, source_vertices, source_tuples[1116], destination_tuples[5365])
    #
    # source_hash_code = source_hash_table_KDTree.data[1116]
    #
    # distance, destination_tuple_index = destination_hash_table_KDTree.query(source_hash_code)

    #
    #
    # def show_match(self, match, figure = None, view='destination'):
    #     if not figure: figure = plt.gcf()
    #     figure.clf()
    #     ax = figure.gca()
    #
    #     #ax.scatter(ps2[:,0],ps2[:,1],c='g',marker = '+')
    #
    #     ax.scatter(match.destination[:,0],match.destination[:,1], marker = '.', facecolors = 'k', edgecolors='k')
    #     ax.scatter(match.transform_source_to_destination[:,0],match.transform_source_to_destination[:,1],c='r',marker = 'x')
    #
    #     destination_basis_index = match.best_image_basis['hashTableBasis']
    #     source_basis_index = match.best_image_basis['testBasis']
    #     ax.scatter(match.destination[destination_basis_index, 0], match.destination[destination_basis_index, 1], marker='.', facecolors='g', edgecolors='g')
    #     ax.scatter(match.transform_source_to_destination[source_basis_index, 0], match.transform_source_to_destination[source_basis_index, 1], c='g',
    #                marker='x')
    #
    #     ax.set_aspect('equal')
    #     ax.set_title('Tile:' + self.tile.name +', File: ' + str(self.files[self.matches.index(match)].relativeFilePath))
    #
    #     if view == 'source':
    #         maxs = np.max(match.transform_source_to_destination, axis=0)
    #         mins = np.min(match.transform_source_to_destination, axis=0)
    #         ax.set_xlim([mins[0], maxs[0]])
    #         ax.set_ylim([mins[1], maxs[1]])
    #     elif view == 'destination':
    #         maxs = np.max(match.destination, axis=0)
    #         mins = np.min(match.destination, axis=0)
    #         ax.set_xlim([mins[0], maxs[0]])
    #         ax.set_ylim([mins[1], maxs[1]])
    #         # ax.set_xlim([0, 31000])
    #         # ax.set_ylim([0, 31000])
    #
    #     name = str(self.files[self.matches.index(match)].relativeFilePath)
    #     print(name)
    #     n = name.replace('\\', '_')
    #
    #     figure.savefig(self.dataPath.joinpath(n + '_raw.pdf'), bbox_inches='tight')
    #     figure.savefig(self.dataPath.joinpath(n + '_raw.png'), bbox_inches='tight', dpi=1000)
    #
    # def loop_through_matches(self, figure=plt.figure()):
    #     plt.ion()
    #     for match in self.matches:
    #         self.show_match(match, figure=figure)
    #         plt.show()
    #         plt.pause(0.001)
    #         input("Press enter to continue")





# destination_coordinate_tuples = [destination[list(t)] for t in destination_tuples]
# source_coordinate_tuples = [source[list(t)] for t in source_tuples]
# source_coordinate_tuple = source_coordinate_tuples[source_tuple_index]
# destination_coordinate_tuple = destination_coordinate_tuples[destination_tuple_index]


#
# mp.icp.scatter_coordinates([source_coordinate_tuple, destination_coordinate_tuple])
#
# def connect_pairs(pairs):
#     for pair in pairs:
#         plt.plot(pair[:,0], pair[:,1], color='r')
#
# def plot_tuple(tuple_coordinates):
#     pair_coordinates = tuple_coordinates[:2]
#     internal_coordinates = tuple_coordinates[2:]
#     plt.figure()
#     center = (pair_coordinates[0] + pair_coordinates[1]) / 2
#     distance = np.linalg.norm(pair_coordinates[0] - pair_coordinates[1])
#     connect_pairs([pair_coordinates])
#     mp.icp.([pair_coordinates, np.atleast_2d(center), internal_coordinates])
#     plt.gca().set_aspect('equal')
#     circle = plt.Circle(center, distance / 2, fill=False)
#     plt.gcf().gca().add_artist(circle)
#     axis_limits = np.array([center-distance/2*1.2, center+distance/2*1.2])
#     plt.xlim(axis_limits[:, 0])
#     plt.ylim(axis_limits[:, 1])
#     plt.xlabel('x')
#     plt.ylabel('y')










#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from pathlib import Path
import cv2
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
import tqdm

from papylio.movie.shading_correction import BaSiC


def BaSiCfun(directory, output_directory, extension='.tif', estimate_darkfield=False, apply_correction=False,
             use_flatfield=None, use_darkfield=None, epsilon=1e-6, l_s=None, l_d=None,
             output_flatfield_filename=None, output_darkfield_filename=None, verbose=False):
    # Create the BaSiC shading correction object
    optimizer = BaSiC(directory, estimate_darkfield=estimate_darkfield, extension=extension, verbose=verbose)

    # Set some optimizer parameters
    optimizer.l_s = l_s
    optimizer.l_d = l_d
    # Prepare the optimization
    optimizer.prepare()

    # Extract the flat and dark fields
    perform_estimation = True
    if use_flatfield is not None:
        img = cv2.imread(use_flatfield, cv2.IMREAD_ANYDEPTH)
        optimizer.set_flatfield(img)
        perform_estimation = False

    if use_darkfield is not None:
        img = cv2.imread(use_darkfield, cv2.IMREAD_ANYDEPTH)
        optimizer.set_flatfield(img)
        perform_estimation = False

    # Perform the estimation
    if perform_estimation:
        optimizer.run()

        # Save the estimated fields (only if the profiles were estimated)
        directory = Path(output_directory)
        directory.mkdir(parents=True, exist_ok=True)
        if output_flatfield_filename is not None:
            flatfield_name = Path(output_flatfield_filename).resolve()
            flatfield_name.parent.mkdir(parents=True, exist_ok=True)
        else:
            flatfield_name = directory / "flatfield.tif"
        if output_darkfield_filename is not None:
            darkfield_name = Path(output_darkfield_filename).resolve()
            darkfield_name.parent.mkdir(parents=True, exist_ok=True)
        else:
            darkfield_name = directory / "darkfield.tif"

        cv2.imwrite(str(flatfield_name), optimizer.flatfield_fullsize.astype(np.float32))
        cv2.imwrite(str(darkfield_name), optimizer.darkfield_fullsize.astype(np.float32))

    # Apply shading correction.
    if apply_correction:
        optimizer.write_images(output_directory, epsilon=epsilon)




def squeeze_channel_from_frames(frames):
    return xr.combine_by_coords(
        [frames.sel(channel=channel).set_index(x='x_pixel', y='y_pixel') for channel in frames.channel])

    # xr.combine_by_coords([frames.sel(channel=channel.index).set_index(x='x_pixel', y='y_pixel') for channel in self.channels])
    #
    # test = frames.assign_coords(channel_index_x=('channel',[0,1]),channel_index_y=('channel',[0,0]))
    # test2 = test.drop('channel').set_index(channel=('channel_index_x','channel_index_y'))
    # test3 = test2.unstack('channel').stack(x_pixel=('channel_index_x','x')).stack(y_pixel=('channel_index_y','y'))
    #
    # test2.unstack('channel').stack(channel=('channel_index_x','channel_index_y'))
    # test2.unstack('channel').stack(x_pixel=('channel_index_x','x'), y_pixel=('channel_index_y','y'))

def spatial_shading_correction(movies, method='BaSiC', illumination_index=0, frame_index=0, estimate_darkfield=True, **kwargs):
    selected_movies = [movie for movie in movies if illumination_index in movie.illumination_indices_in_movie]
    frame_with_illumination = [np.where(movie.illumination_index_per_frame == illumination_index)[0][frame_index] for movie in selected_movies]
    # frames = xr.concat([movie.read_frames([frame], apply_corrections=False, xarray=True, flatten_channels=False)
    #                     for movie, frame in tqdm.tqdm(zip(selected_movies, first_frame_with_illumination))], dim='frame')
    # frames = frames.reset_index('frame', drop=True)
    frames = np.vstack([movie.read_frames([frame], apply_corrections=False, xarray=False, flatten_channels=False)
                        for movie, frame in tqdm.tqdm(zip(selected_movies, frame_with_illumination),
                                            'Read frames', len(selected_movies))])

    flatfield = np.ones_like(frames[0], dtype=float)
    if estimate_darkfield:
        darkfield = np.zeros_like(frames[0], dtype=float)
    else:
        darkfield = selected_movies[0].corrections.darkfield_correction.sel(illumination=illumination_index).values
        frames = frames - darkfield[None, :, :, :]

    if method == 'BaSiC':
        for channel_index in movies[0].channel_indices:
            optimizer = BaSiC(frames[:, channel_index, :, :], estimate_darkfield=estimate_darkfield, extension=None, verbose=True, **kwargs)
            optimizer.prepare()
            optimizer.run()
            flatfield[channel_index,:,:] = optimizer.flatfield_fullsize
            if estimate_darkfield:
                darkfield[channel_index,:,:] = optimizer.darkfield_fullsize
    elif method == 'average':
        flatfield = frames.mean(axis=0)
        flatfield = flatfield / flatfield.mean(axis=(-2,-1), keepdims=True)

    # flatfield = squeeze_channel_from_frames(flatfield)
    # darkfield = squeeze_channel_from_frames(darkfield)

    flatfield = movies[0].flatten_channels(flatfield)
    darkfield = movies[0].flatten_channels(darkfield)

    return darkfield, flatfield

#
# movies = files_green_laser[0::25].movie
# darkfield, flatfield = spatial_shading_correction(movies)
#
# import tifffile
# save_path = Path(r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20220602 - Objective-type TIRF (BN)')
# tifffile.imwrite(save_path.joinpath('flatfield.tif'), flatfield)
# tifffile.imwrite(save_path.joinpath('darkfield.tif'), darkfield)

# plt.figure()
# plt.imshow(flatfield)
#
# plt.figure()
# plt.imshow(darkfield)

#
# frames = files_green_laser[0].movie.read_frames_raw()
# frames = squeeze_channel_from_frames(frames)

if __name__ == '__main__':
    import tifffile
    from pathlib2 import Path
    import numpy as np

    pth = Path(r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20220602 - Objective-type TIRF (BN)\Analysis\Test')

    img_stack = tifffile.imread(pth.joinpath('TIRF 561 0001.tif'))
    img_stack = np.rot90(img_stack, 1, axes=(1,2))
    flatfield = tifffile.imread(pth.joinpath('flatfield.tif'))
    darkfield = tifffile.imread(pth.joinpath('darkfield.tif'))

    from papylio.movie.shading_correction import get_photobleach

    import time
    start = time.time()
    c1 = get_photobleach(img_stack[:,:,0:256], flatfield[:,0:256], darkfield[:,0:256], size=(32,64))
    print(time.time()-start)
    start = time.time()
    c2 = get_photobleach(img_stack[:,:,256:512], flatfield[:,256:512], darkfield[:,256:512], size=(32,64))
    print(time.time()-start)

    import matplotlib.pyplot as plt
    plt.figure()
    plt.plot(c1.T/c1.max())
    plt.plot(c1.T*c2.max()/c2.T)
    plt.plot(c2.T-c1.T*c2.max()/c2.T)
    plt.plot(c2.max()/c2.T)
    plt.plot(c2.T/c2.max())
    plt.plot(test2/test2.max())




    from papylio.movie.shading_correction import get_photobleach
    import time
    start = time.time()
    test = get_photobleach(frames.values, flatfield, darkfield=darkfield)
    print(time.time()-start)

    import scipy.ndimage
    start = time.time()
    # test2 = np.array([scipy.ndimage.minimum_filter(((frame)), size=15, mode='wrap').sum() for frame in img_stack])
    test2 = np.array([scipy.ndimage.gaussian_filter(((frame-darkfield)/flatfield)[:,0:256], sigma=50, mode='wrap').mean() for frame in img_stack])
    test2 = np.array([scipy.ndimage.minimum_filter(((frame-darkfield)/flatfield)[:,0:256], size=15, mode='wrap').mean() for frame in img_stack])
    # test2 = np.array([frame.sum() for frame in frames])
    # test2 = np.array([scipy.ndimage.minimum_filter(frame, size=15, mode='wrap').sum() for frame in img_stack])
    print(time.time()-start)

    test = test.squeeze()
    test = test/test.sum()
    test2 = test2/test2.sum()

    plt.figure()
    plt.plot(test)
    plt.plot(test2)








# -*- coding: utf-8 -*-
"""
Created on Sat Apr 27 23:12:58 2019

@author: iason
"""
if __name__ == '__main__':
    import sys
    from pathlib import Path
    p = Path(__file__).parents[2]  # include two levels up directory where the papylio package lies
    sys.path.insert(0, str(p))

import numpy as np

from papylio.experiment import Experiment
import pandas as pd
import os
import itertools
import matplotlib.pyplot as plt
import matplotlib.widgets
import seaborn as sns
from pathlib import Path, WindowsPath
import functools
#plt.rcParams['toolbar'] = 'toolmanager'
# from matplotlib.backend_tools import ToolBase


class InteractivePlot(object):
    def __init__(self, molecules, canvas=None):

        self.molecules = molecules
        self.mol_indx = 0  #From which molecule to start the analysis
        self.canvas = canvas


    @property
    def file(self):
        return self.molecules[self.mol_indx].file
    @property
    def time(self):
        return self.molecules[self.mol_indx].file.time

    def plot_initialize(self):
        plt.ion() # Possibly we should fix this in another way [IS: 18-11-2019]

        sns.set(style="dark")
        sns.set_color_codes()
        plt.style.use('dark_background')
        if self.canvas is None:
            self.fig, self.axes = plt.subplots(2, 1, sharex=True, figsize=(10,5))
        else:
            self.fig = self.canvas.figure
            self.axes = self.fig.subplots(2, 1, sharex=True)


        self.axes[0].set_ylim((-20, 500))  # Set default intensity limits
        self.axes[0].set_ylabel("intensity (a.u)\n")
        self.axes[1].set_ylim((-0.1,1.1))  # Set default fret limits
        self.axes[1].set_xlabel("time (s)")
        self.axes[1].set_ylabel("FRET\n")
        self.axes[1].set_xlim((-10, self.time[-1]+10))  # Set default x limits dependent on measurement time

        plt.subplots_adjust(bottom=0.23)

        # Create the axes for the widgets
        self.rax = plt.axes([0.91, 0.71, 0.08, 0.2])


        self.axcheckfret = plt.axes([0.91, 0.35, 0.08, 0.08])
        self.axcorred = plt.axes([0.95, 0.6, 0.028, 0.06])
        self.axcorgreen = plt.axes([0.95, 0.53, 0.028, 0.06])
        self.axcorrfretI = plt.axes([0.95, 0.3, 0.028, 0.06])
        self.axthrsliders = [plt.axes([0.26, 0.072, 0.10, 0.03]),
                             plt.axes([0.26, 0.033, 0.10, 0.03])]
        # Create the buttons
        self.axthresb = plt.axes([0.1, 0.03, 0.13, 0.062])  # Button to calculate dwell times by thresholding
        self.axrejb = plt.axes([0.41, 0.03, 0.07, 0.062])   # Button to reject calculated dwell times by thresholding
        self.axclearb = plt.axes([0.51, 0.03, 0.11, 0.062]) # Button to clear the clicked points (clears vlines)
        self.axthrowb = plt.axes([0.64, 0.03, 0.11, 0.062]) # Button to throw away already calculated dwell times and de-select molecule
        self.axconclb = plt.axes([0.77, 0.03, 0.15, 0.062]) # Button to conlcude analysis by saving all the calculated steps and metadata

        self.axnextb = plt.axes([0.162, 0.90, 0.065, 0.062])  # Buttons to cycle through molecules
        self.axprevb = plt.axes([0.075, 0.90, 0.08, 0.062])

        self.axsavetraceb = plt.axes([0.26, 0.90, 0.065, 0.062])

        self.axgotomolecule = plt.axes([0.50, 0.89, 0.05, 0.062])
        self.axplotselected = plt.axes([0.63, 0.90, 0.065, 0.062])
        self.axsaveselectedb = plt.axes([0.77, 0.90, 0.065, 0.062])

        self.axtotal = plt.axes([0.865, 0.90, 0.065,  0.062])
        [ax.set_frame_on(False) for ax in self.fig.get_axes()[2:]]


        #  Radiobutton to select red or green
        self.radio = matplotlib.widgets.RadioButtons(self.rax, ("red", "green", 'total'))
        self.radio.circles[0].set_color("r")
        for circle in self.radio.circles: # adjust radius here. The default is 0.05
            circle.set_radius(0.07)
        self.radio.on_clicked(self.radio_manage)

        #  Connect clicking to draw lines class
        self.draw = Draw_lines(self.fig, self.radio)

        # Create the control buttons
        bp = {'color': 'black', 'hovercolor': 'gray'}
        self.bauto = matplotlib.widgets.Button(self.axthresb,'autoThreshold' , **bp)
        self.bauto.on_clicked(self.autoThreshold_plot)
        self.brejauto = matplotlib.widgets.Button(self.axrejb,'reject' , **bp)
        self.brejauto.on_clicked(self.auto_reject)
        self.bclear = matplotlib.widgets.Button(self.axclearb,'clear clicks' , **bp)
        self.bclear.on_clicked(self.draw.clear_all)
        self.bthrow = matplotlib.widgets.Button(self.axthrowb,'throw away' , **bp)
        self.bthrow.on_clicked(self.throw_away)
        self.bconcl = matplotlib.widgets.Button(self.axconclb,'conclude analysis' , **bp)
        self.bconcl.on_clicked(self.conclude_analysis)
        self.bnext = matplotlib.widgets.Button(self.axnextb,'Next' , **bp)
        self.bnext.on_clicked(self.save_molecule)
        self.bprev = matplotlib.widgets.Button(self.axprevb,'Previous' , **bp)
        self.bprev.on_clicked(self.save_molecule)

        self.bsavetrace = matplotlib.widgets.Button(self.axsavetraceb,'save trace' , **bp)
        self.bsavetrace.on_clicked(self.save_trace)

        # an entry box to go to desired molecule
        gotodict = {'initial': str(1), 'color':'k', 'hovercolor': 'k', 'label_pad':.2}
        self.gotoentry = matplotlib.widgets.TextBox(self.axgotomolecule,
                                                    'Go To:', **gotodict)

        # a checkbutton to whether plot only the selected molecules
        self.checkbplotselected = matplotlib.widgets.CheckButtons(self.axplotselected,
                                                                  ["plot selected"],
                                                                  actives=[False])

        # A button to save selected molecules
        self.bsaveselected = matplotlib.widgets.Button(self.axsaveselectedb,'save selected' , **bp)
        self.bsaveselected.on_clicked(self.save_selected)

        #  A checkbutton for whether to display the total intensity
        self.checkbtotal = matplotlib.widgets.CheckButtons(self.axtotal, ["show total"],
                                                          actives=[False])

        #  A checkbutton for fret autothreshold dwell-time calculation
        self.checkbfret = matplotlib.widgets.CheckButtons(self.axcheckfret, ["E fret"],
                                                          actives=[False])

        #  Remove the x lines from the checkbuttons
        for chbutton in [self.checkbplotselected, self.checkbtotal, self.checkbfret]:
            chbutton.rectangles[0].set_color("black")
            chbutton.rectangles[0].set_height( 0.2)
            [line.remove() for line in chbutton.lines[0]]

        #  Entryboxes for offset corrections
        corrdict = {'initial': str(0), 'color':'k', 'hovercolor': 'k', 'label_pad':.2}
        corrlabels = [r'$I_{R_{off}}$', r'$I_{G_{off}}$', r'$I_{min}$']
        corraxes = [self.axcorred, self.axcorgreen, self.axcorrfretI]
        self.correntries = [matplotlib.widgets.TextBox(ax, label, **corrdict)
                            for ax, label in zip(corraxes, corrlabels)]

        #  Sliders for assigning the threshold
        self.thrsliders = []
        self.thrsliders.append(matplotlib.widgets.Slider(self.axthrsliders[0],
                                                         label=r"$I_R$", valmin=0,
                                                         valmax=200, valinit=50,
                                                         valfmt="%i", color="r"))
        self.thrsliders.append(matplotlib.widgets.Slider(self.axthrsliders[1],
                                                         label=r"$E$", valmin=0,
                                                         valfmt="%.2f", valinit=0.5,
                                                         color="b", valmax=1.0))
        [slider.vline.remove() for slider in self.thrsliders]  # remove the default vertical lines showing the initial value

        self.connect_events_to_canvas()

        # remove keyboard shortcuts from default matplotlib
        try:
            plt.rcParams['keymap.home'].remove('r')
            plt.rcParams['keymap.grid'].remove('g')
        except ValueError:
            pass


    def connect_events_to_canvas(self):
        self.fig.canvas.mpl_connect('button_press_event', self.draw.onclick)
        self.fig.canvas.mpl_connect('key_press_event', self.key_bind)
        self.fig.canvas.mpl_connect('axes_leave_event', functools.partial(self.change_axis, 'leave'))
        self.fig.canvas.mpl_connect('axes_enter_event', functools.partial(self.change_axis, 'enter'))
        [entry.on_submit(lambda _: self.plot_molecule()) for entry in self.correntries]
        [slider.on_changed(functools.partial(self.change_slider, slider)) for slider in self.thrsliders]
        self.checkbplotselected.on_clicked(self.checkbutton_color)
        self.gotoentry.on_submit(self.go_to_molecule)
        self.checkbtotal.on_clicked(self.check_total)
        self.checkbfret.on_clicked(self.checkbutton_color)


    def plot_molecule(self, draw_plot=True):
        sns.set(style="dark")
        sns.set_color_codes()
        plt.style.use('dark_background')
        # clear the appropriate lines from axes first
        [ax.lines.clear() for ax in self.fig.get_axes()[:2]]
        # find the current molecule instance
        self.mol = self.molecules[self.mol_indx]
        print(f'mol. coordinates: {self.mol.coordinates}')
        self.fig.canvas.set_window_title(f'Dataset: {self.file.name}')

        # Check if molecule is selected, This will also set the title of the axis
        if self.mol.isSelected:
            self.select_molecule(toggle=False)
        else:
            self.select_molecule(toggle=False, deselect=True)
        #load saved steps if available
        self.load_from_Molecule()

        # load kon if existing or assign a False 3x3 boolean
        self.prev_mol = self.molecules[self.mol_indx - 1]
        if all(kon is None for kon in [self.mol.kon_boolean, self.prev_mol.kon_boolean]):
            self.kon = np.zeros((4,3), dtype=bool)
        elif self.mol.kon_boolean is  None:
            self.kon = np.copy(self.prev_mol.kon_boolean)  # if no kon is defined for current molecule
        else:
            self.kon = self.mol.kon_boolean

        # update the edge color from self.kon:
        self.load_edges(load_fret=True)

        self.Iroff, self.Igoff, self.Imin = [float(c.text) for c in self.correntries]

        self.red = self.mol.I(1, Ioff=self.Iroff)
        self.green = self.mol.I(0, Ioff=self.Igoff)
        self.total = self.red + self.green


        self.fret = self.mol.E(Imin=self.Imin, Iroff=self.Iroff, Igoff=self.Igoff)

        if not draw_plot:
            return

        #  if draw_plot is true
        # The order in which the traces are plotted matters later for the order of the axes lines.
        # it should always follow the order of the radio button
        self.lred = self.axes[0].plot(self.time, self.red, "r", lw=.75)[0]
        self.lgreen = self.axes[0].plot(self.time, self.green, "g", lw=.75)[0]
        self.ltotal = self.axes[0].plot(self.time, self.total, "bisque", lw=.65,
                               zorder=-1, visible=self.checkbtotal.get_status()[0])[0]

        Imax = np.max(self.red + self.green)
        self.axes[0].set_ylim((self.axes[0].get_ylim()[0], Imax))

        self.axes[1].plot(self.time, self.fret, "b", lw=.75)

        # vertical lines to indicate the threshold in the two axes
        self.slidel = [ax.axhline(0, lw=1, ls=":", zorder=3, visible=False) for ax in self.axes]
        #  Creat cursor particular to the molelcule and connect it to mouse movement event
        self.cursors = []
        cursor_kws = {'useblit': True, 'color': 'white', 'ls': "--", 'lw': 1, 'alpha': 0.5}
        self.cursors.append(matplotlib.widgets.Cursor(self.axes[0], **cursor_kws))
        self.cursors.append(matplotlib.widgets.Cursor(self.axes[1], **cursor_kws))

        self.fig.canvas.draw()

    def change_axis(self, event_type, event):
        ax = event.inaxes
        if event_type == 'enter':
            if ax == self.axes[1]:
                self.fret_edge_lock = False
            elif ax in self.axthrsliders:
                indx = int(ax == self.axthrsliders[1])  # gives 0 if ax is upper (I) plot, 1 if ax is lower (E) plot
                self.slidel[indx].set_ydata(self.thrsliders[indx].val)
                self.slidel[indx].set_visible(True)
                self.fig.canvas.draw()
        elif event_type in ['leave', 'slider_change']:
            if ax == self.axes[1]:
                self.fret_edge_lock = True
            elif ax in self.axthrsliders:
                indx = int(ax == self.axthrsliders[1])  # gives 0 if ax is upper (I) plot, 1 if ax is lower (E) plot
                self.slidel[indx].set_visible(False)
                self.fig.canvas.draw_idle()

    def change_slider(self, slider, cid):
        indx = int(slider == self.thrsliders[1])  # # gives 0 if slider is I slider, 1 if slider E slider
        self.slidel[indx].set_ydata(self.thrsliders[indx].val)
        self.slidel[indx].set_visible(True)
        self.fig.canvas.draw_idle()


    def key_bind(self, event):

        k = event.key
        if k == 'a': self.autoThreshold_plot(event, find_all=False)
        if k == 'ctrl+a': self.autoThreshold_plot(event, find_all=True)
        elif k in ['w', 'e']: self.save_molecule(event, move=True)
        elif k == 'z': self.auto_reject(event)
        elif k == 'c': self.draw.clear_all(event)
        elif k in [',', '.', '/']: self.select_edge(k)
        elif k == ' ': self.select_molecule(toggle=True)
        elif k == 'r': self.radio_manage('red')
        elif k == 'g': self.radio_manage('green')
        elif k == 't': self.radio_manage('total')
        elif k == 'f':
            self.checkbfret.set_active(0)
            self.checkbutton_color('E fret')
        elif k == 'i':
            self.checkbtotal.set_active(0)
            self.check_total('show total')

        elif k == 'x': self.throw_away(event)
        elif k == 'l': self.conclude_analysis()
        elif k == '[': self.draw.select_starttrace(event)
        elif k == ']': self.draw.select_endtrace(event, self.time[-1])

        self.fig.canvas.draw_idle()


    def load_from_Molecule(self):
        if self.mol.steps is None:
            return
        else:
            s = self.mol.steps
            [self.axes[0].axvline(f, zorder=0, lw=0.4, c='red', label="saved r")
             for f in s.time[s.trace == 'red'].values]
            [self.axes[0].axvline(f, zorder=0, lw=0.4, c='lime', label="saved g")
             for f in s.time[s.trace == 'green'].values]
            [self.axes[0].axvline(f, zorder=0, lw=0.4, c='gold', label="saved t")
             for f in s.time[s.trace == 'total'].values]
            [self.axes[1].axvline(f, zorder=0, lw=0.4, c='aqua', label="saved E")
             for f in s.time[s.trace == 'E'].values]

    def select_molecule(self, toggle=True, deselect=False):
        if toggle:
            self.mol.isSelected = not self.mol.isSelected
        elif deselect:
            self.mol.isSelected = False
        else:
            self.mol.isSelected = True

        title = f'Molecule: {self.molecules.index(self.mol)+1} /{len(self.molecules)} '
        title += f'({self.mol.index + 1} /{len(self.mol.file.molecules)} in {self.mol.file.name})'
        title += '  (S)'*(self.mol.isSelected)
        rgba = matplotlib.colors.to_rgba
        c = rgba('g')*self.mol.isSelected + rgba('w')*(not self.mol.isSelected)
        self.fig.suptitle(title, color=c, fontsize=10)
        self.fig.canvas.draw_idle()

    def throw_away(self, event):
        if self.mol.steps is not None:
            lines = self.axes[0].get_lines() + self.axes[1].get_lines()
            # select only the vertical lines with corresponding labels
            lines = [l for l in lines if l.get_label().split()[0] \
                     in ['man','thres','saved']]
            # check if the lines are inside the current zoom level
            lines = [l for l in lines if \
                     l.get_xdata()[0] > self.axes[0].get_xlim()[0] and \
                         l.get_xdata()[0] < self.axes[0].get_xlim()[1]]

            # remove the corresponding data from molecule.steps
            for l in lines:
                steps = self.mol.steps.time.values
                xdat = l.get_xdata()[0]
                indx = next(i for i, _ in enumerate(steps) \
                                if np.isclose(_, xdat, 10**-4))
                self.mol.steps.drop(indx, inplace=True)
                self.mol.steps.reset_index(drop=True, inplace=True)
                print(self.mol.steps)
            # set mol.steps to None if it ends up being empty and deselect it
            if self.mol.steps.empty:
                self.mol.steps = None
                self.select_molecule(toggle=False, deselect=True)

            # remove the selected lines
            [print(f'line at {l.get_xdata()[0]} removed') for l in lines]
            [l.remove() for l in lines]

            self.fig.canvas.draw_idle()


    def save_molecule(self, event=None, move=True, draw=True):

#       Assume acceptance of automatically found and manually selected dwell times
        lines = self.axes[0].get_lines() + self.axes[1].get_lines()
        lines = [l for l in lines if l.get_label().split()[0] in ["man", "thres"]]
        self.mol.kon_boolean = self.kon
        if lines:
            if len(lines) % 2 != 0:
                print(f'Found an odd number of steps. Molecule {self.mol.index} not added')
                return
            if self.mol.steps is None:
                print('lines', len(lines))
                self.mol.steps = pd.DataFrame(columns=['time', 'trace', 'state',
                                                       'method','thres','kon',
                                                       'Iroff', 'Igoff', 'Imin'],
                                              dtype=object)
            # Molecule is automatically selected if steps are indicated
            # self.mol.isSelected = True
            # turn the kon matrix into a flat string
            kon = [f'{int(i)}' for i in self.mol.kon_boolean.flatten()]
            kon = ''.join(kon)

            for l in lines:
                method = l.get_label().split()[0]
                thres = "N/A"*(method=='man') + str(self.thrsliders[0].val)*(method =='thres')

                d = {'time': max(l.get_xdata()[0], 0), 'trace': l.get_label().split()[1],
                     'state': 1, 'method': method, 'thres': thres, 'kon': kon,
                      'Iroff': self.Iroff, 'Igoff': self.Igoff, 'Imin': self.Imin,
                      'isSelected': bool(self.mol.isSelected)}
                self.mol.steps = self.mol.steps.append(d, ignore_index=True)
            self.mol.steps.drop_duplicates(inplace=True)

           #Sort the timepoints first by trace type and then in ascending order
            self.mol.steps = self.mol.steps.sort_values(by=['trace', 'time'])
            self.mol.steps.reset_index(inplace=True, drop=True)

        if move:
            self.move_molecule(event, draw)


    def move_molecule(self, event, draw):
        if event.inaxes == self.axnextb or event.key in ['e']:
            self.mol_indx += 1
            if self.mol_indx >= len(self.molecules) \
                or self.mol_indx <= -len(self.molecules):
                self.mol_indx = 0

        elif event.inaxes == self.axprevb or event.key in ['w']:
            self.mol_indx -= 1
            if self.mol_indx >= len(self.molecules) \
                or self.mol_indx <= -len(self.molecules):
                self.mol_indx = 0

        # if the plot_selected checkbutton is active move again until a selected
        # molecule is found
        if self.checkbplotselected.get_status()[0]:
            mol = self.molecules[self.mol_indx]
            if not mol.isSelected:
                self.move_molecule(event, draw)
            else:
                self.plot_molecule(draw_plot=draw)
        else:

            self.plot_molecule(draw_plot=draw)

    def conclude_analysis(self, event=None, save=True, filename=None):
        # Save current molecule if it was analyzed
        self.save_molecule(move=False)

        if filename is None:
            filename = f'{self.file.relativeFilePath}_steps_data.xlsx'

        if save:
            self.file.savetoExcel(filename=filename)
            # save also the selected molecules
            self.save_selected()


    def autoThreshold_plot(self, event=None, find_all=False):
        self.auto_reject()
        #  Find the steps for the checked buttons
        sel = self.radio.value_selected
        c = 'red'*(sel=='red') + 'lime'*(sel=='green')\
                    + 'gold'*(sel=='total') + 'aqua'*(sel=='E')
        trace = self.red*bool(sel == 'red') + self.green*bool(sel == 'green') \
                + self.total*bool(sel == 'total')  # Select trace data for red  or green
        # get the color indices for the current zoom level
        i_left = find_nearest(self.time, self.axes[0].get_xlim()[0])
        i_right = find_nearest(self.time, self.axes[0].get_xlim()[1])
        # print(i_left, i_right)

        mask = np.zeros(trace.size)
        mask[i_left: i_right] = 1
        trace = np.multiply(trace, mask)

        steps = self.mol.find_steps(trace, threshold=self.thrsliders[0].val)
        # print(steps)
        l_props = {"lw": 0.4, "zorder": 5,'color':c, "label": "thres "+sel}
        [self.axes[0].axvline(s*self.file.exposure_time, **l_props) \
         for s in steps["frames"]]
        if self.checkbfret.get_status()[0]:
            fret = self.fret[i_left: i_right]
            steps = self.mol.find_steps(fret, threshold=self.thrsliders[1].val)
            l_props = {"lw": 0.75, "zorder": 5, "label": "thres E"}
            [self.axes[1].axvline(s*self.mol.file.exposure_time, **l_props) \
             for s in steps["frames"]]
        self.fig.canvas.draw()
        if find_all:
            for mol in self.molecules:
                self.autoThreshold_plot(find_all=False)
                print(f'Analyzed mol {self.mol.index} /{len(self.molecules)}')
                e = matplotlib.backend_bases.KeyEvent('key_press_event',
                                                      self.fig.canvas, 'e')
                if mol != self.molecules[-1]:
                    self.save_molecule(event=e, move=True, draw=False)
                elif mol == self.molecules[-1]:
                    self.conclude_analysis()
                    return

    def auto_reject(self, event=None):
        for ax in self.axes:
            lines = ax.get_lines()
            lines = [l for l in lines if l.get_label().split()[0] == 'thres']
            lines = [l for l in lines if \
                     l.get_xdata()[0] > self.axes[0].get_xlim()[0] and \
                         l.get_xdata()[0] < self.axes[0].get_xlim()[1]]
            [l.remove() for l in lines]
            self.fig.canvas.draw_idle()


    def radio_manage(self, label):
        def update_slider(color, label):
            s = self.thrsliders[0]
            s.poly.set_color(color); s.label.set(text=label)

        indx = [l.get_text() for l in self.radio.labels].index(label)
        color = 'r'*(label=='red') + 'g'*(label=='green') + 'bisque'*(label=='total')

        lines =  [self.lred, self.lgreen, self.ltotal]
        [l.set_zorder(-lines.index(l)) for l in lines if lines.index(l) != indx ]
#        print(indx, color)
        lines[indx].set_zorder(10)
        self.radio.circles[indx].set_color(color)
        [c.set_color('black') for c in self.radio.circles if self.radio.circles.index(c) != indx]
        update_slider(color, r"$I_G$"*(label=='green') + \
                                r"$I_R$"*(label=='red') + \
                                r"$I_T$"*(label=='total') )

        self.radio.value_selected = label
        # Check the edge colors and set to white if not selected color
        # sel = self.radio.value_selected
        selcol = matplotlib.colors.to_rgba(color)
        spcol = [self.axes[0].spines[s].get_edgecolor() for s in ['left','bottom','right']]
        if selcol not in spcol:
            [self.axes[0].spines[s].set_color('white') for s in ['left','bottom','right']]

        self.load_edges()

    def load_edges(self, load_fret=False):  # loads edge color from kon array
        label = self.radio.value_selected
        indx = [l.get_text() for l in self.radio.labels].index(label)
        color = 'r'*(label=='red') + 'g'*(label=='green') + 'sandybrown'*(label=='total')

        kons = [self.kon[indx]]
        colors = [color]
        if load_fret: kons.append(self.kon[3]) ;colors.append('blueviolet')

        for i, kon in enumerate(kons):
            selected_sides = list(itertools.compress(['left','bottom','right'], kon))
            unselected_sides = list(itertools.compress(['left','bottom','right'], np.invert(kon)))
            [self.axes[i].spines[s].set_color(colors[i]) for s in selected_sides]
            [self.axes[i].spines[s].set_color('white') for s in unselected_sides]

        self.fig.canvas.draw_idle()

    def select_edge(self, key):
        if self.fret_edge_lock:
            ax = self.axes[0]
            label = self.radio.value_selected  # get the selected color of the radiobutton
            color = 'r'*(label=='red') + 'g'*(label=='green') + 'sandybrown'*(label=='total')
        elif not self.fret_edge_lock:
            ax = self.axes[1]
            color = 'blueviolet'  # this refers to the fret color

        side = 'left'*(key == ',')  + 'bottom'*(key == '.') + 'right'*(key == '/')

        spcolor = ax.spines[side].get_edgecolor()
        selcol, w = matplotlib.colors.to_rgba(color), matplotlib.colors.to_rgba('white')
        c = selcol*(spcolor == w) + w*(spcolor == selcol)
        ax.spines[side].set_color(c)

        self.update_kon(color, selcol, side, ax)

    def update_kon(self, sel=None, selcol=None, side=None, ax=None):
        i = ['r', 'g', 'sandybrown', 'blueviolet'].index(sel)  # These are the colors of the sides. blueviolet refers to fret
        j = ['left', 'bottom', 'right'].index(side)
        self.kon[i][j] = (ax.spines[side].get_edgecolor() == selcol)


    def check_total(self, label):
        if self.checkbtotal.get_status()[0]:
            self.ltotal.set_visible(True)
        else:
            self.ltotal.set_visible(False)
        self.checkbutton_color(label)


    def checkbutton_color(self, label):  # changes the color of the fret checkbutton. Purely for aesthetics
        if label == 'E fret':
            chbutton = self.checkbfret ; c = 'b'
        elif label == 'show total':
            chbutton = self.checkbtotal ; c = 'bisque'
        elif label == 'plot selected':
            chbutton =  self.checkbplotselected ; c = 'gold'

        if chbutton.get_status()[0]:
            chbutton.rectangles[0].set_color(c)
        elif not chbutton.get_status()[0]:
            chbutton.rectangles[0].set_color("black")
        self.fig.canvas.draw_idle()

    def save_trace(self, event=None):

        Ioff = [self.Igoff, self.Iroff]
        self.mol.plot(xlim=self.axes[0].get_xlim(),
                      ylim=self.axes[0].get_ylim(),
                      Ioff=Ioff, Imin=self.Imin,
                      Iroff=self.Iroff, Igoff=self.Igoff, save=True)

    def save_selected(self, event=None):
        selected_molecules = []
        file = self.molecules[0].file
        for mol in self.molecules:

            if mol.isSelected:
                current_file = mol.file
                if current_file == file:
                    selected_molecules.append(mol.index+1)
                else:
                    if bool(selected_molecules):
                        np.savetxt(f'{file.relativeFilePath}_selected_molecules.txt',
                                   np.array(selected_molecules, dtype=int))
                        file = current_file
                        selected_molecules = []
        if bool(selected_molecules):

            np.savetxt(f'{file.relativeFilePath}_selected_molecules.txt',
                                   np.array(selected_molecules, dtype=int))
        else:
            print('no molecules selected')

    def go_to_molecule(self, event=None):
        self.mol_indx = int(self.gotoentry.text) - 1
        self.plot_molecule()


class Draw_lines(object):
    def __init__(self, fig, iplot_radio):
        self.lines = []
        self.fig = fig
        self.radio = iplot_radio  # The InteractivePlot instance

    def onclick(self, event):
        if self.fig.canvas.manager.toolbar.mode != '':  # self.fig.canvas.manager.toolmanager.active_toggle["default"] is not None:
            return
        if event.inaxes is None:
            return
        ax = event.inaxes
        if event.button == 1:
            if ax == self.fig.get_axes()[0] or ax == self.fig.get_axes()[1]:
                sel = self.radio.value_selected*(ax == self.fig.get_axes()[0])
#                print(sel)
                sel = sel + "E"*(ax == self.fig.get_axes()[1])
                c = 'red'*(sel=='red') + 'lime'*(sel=='green')\
                    + 'gold'*(sel=='total') + 'aqua'*(sel=='E')
                l = ax.axvline(x=event.xdata, zorder=0, lw=0.4, c=c, label="man "+sel)
                self.lines.append(l)

        if event.button == 3 and self.lines != []:
            self.lines.pop().remove()
        self.fig.canvas.draw_idle()

    def select_starttrace(self, event):
        if event.inaxes is None:
            return
        ax = event.inaxes
        sel = self.radio.value_selected
        l = ax.axvline(0, zorder=0, lw=0.65, c='yellow', label="man "+sel)
        self.lines.append(l)
        self.fig.canvas.draw_idle()

    def select_endtrace(self, event,endtime):
        if event.inaxes is None:
            return
        ax = event.inaxes
        sel = self.radio.value_selected
        l = ax.axvline(endtime, zorder=0, lw=0.65, c='yellow', label="man "+sel)
        self.lines.append(l)
        self.fig.canvas.draw_idle()

    def clear_all(self, event):
        while self.lines:
            self.lines.pop().remove()
        self.fig.canvas.draw_idle()

def find_nearest(array, value):
    # array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return int(idx)


if __name__ == '__main__':

    # Just as a working example of how the interactive plot whould be called. Here an example dataset is included inside the traces folder
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    p = Path(__file__).parents[2]

    # main_path = Path(str(p) + '/traces')
    main_path = os.path.join(p, WindowsPath('traces'))  # This path cannot be found strangely
    main_path = r'C:\Users\iason\Desktop\traceanalysis\trace_analysis\traces\test_data\DNA04'
    exp = Experiment(main_path)
    file = exp.files[0]
    molecules = file.molecules
    i = InteractivePlot(molecules)
    i.plot_initialize()
    i.plot_molecule()
    plt.show()


#self.fig.canvas.manager.toolmanager.add_tool('Next', NextTool)
#self.fig.canvas.manager.toolbar.add_tool('Next', 'foo')
#class NextTool(ToolBase, InteractivePlot):
#    '''Go to next molecule'''
#    default_keymap = 'enter, right'
#    description = 'Next Molecule 1'
#
#    def trigger(self, *args, **kwargs):
#        pass
#        InteractivePlot.__init__(InteractivePlot, self.file)
#        print(self.mol_indx
#              )
#        InteractivePlot.plot_setup(InteractivePlot)
#        print(InteractivePlot.mol)


# -*- coding: utf-8 -*-
"""
Created on Tue May 28 12:05:13 2019

@author: iason
"""

import concurrent.futures
import time
import numpy as np
import pandas as pd


def find_mol_dwells(mol, trace='red'):
    max_time = mol.file.time[-1]
    exp_time = mol.file.exposure_time

    times = mol.steps.time.values[mol.steps.trace == trace]
    try:
        times1 = times.reshape((int(times.size/2), 2))
    except ValueError:
        return

    offtimes = np.diff(times1, axis=1).flatten()

    labels = []
    for i, d in enumerate(offtimes):
        lab = 'm'
        if times[0] < 1 and i == 0:  # first loop
            lab = 'l'
        if max_time - times[-1] < 0.1 and i == len(offtimes) - 1:  # last loop
            lab = 'r'
        labels.append(lab)
    offtimes = pd.DataFrame({'offtime': offtimes, 'side': labels})

    # Calculate the on times
    ontimes = []
    labels = []
    if times[0] > exp_time:  # append the left kon if it exists
        ontimes.append(times[0])
        labels.append('l')

    for i in range(2, times.size, 2):
        ontimes.append(times[i] - times[i-1])
        labels.append('m')

    if max_time - times[-1] > exp_time:  # append the right kon if it exists
        ontimes.append(max_time - times[-1])
        labels.append('r')

    ontimes = pd.DataFrame({'ontime': ontimes,
                            'onside': labels,
                            'order': np.arange(0, len(ontimes))})

  # Calculate the average FRET for each dwell

    avg_fret = []
    Icheck = int((mol.steps.Imin.tail(1)== mol.steps.Imin[0]) &
                  (mol.steps.Iroff.tail(1) == mol.steps.Iroff[0]) &
                  (mol.steps.Igoff.tail(1) == mol.steps.Igoff[0]))
    if Icheck == 1:  #  check if thresholds the same for each dwell of the molecule
        fret = mol.E(Imin=mol.steps.Imin[0], Iroff=mol.steps.Iroff[0], Igoff=mol.steps.Igoff[0])
    else:
        print(f'Ioffsets are not equal for molecule:{i+1}')
        fret = []

    for ii in range(0, len(times)):
        if ii % 2 != 0:
            istart = int(round(times[ii-1]/exp_time))
            iend = int(round(times[ii]/exp_time))
            a_fret = round(np.mean(fret[istart:iend]), 2)
            if (a_fret <= 1 and a_fret >= 0):
                avg_fret.append(a_fret)
            else:
                avg_fret.append(0)
                print(f'FRET corrupted for molecule:{i+1}')
        # else:
        #     avg_fret.append(None)

    avgFRET = pd.DataFrame({'avrgFRET': avg_fret})

    return pd.concat([offtimes, ontimes, avgFRET], axis=1)


def process_molecule(mol):
    traces_unique = pd.unique(mol.steps.trace.values)
    results = []
    for trace in traces_unique:

        results.append(find_mol_dwells(mol, trace=trace))

    result = pd.concat(results, axis=0, ignore_index=True)
    mol.steps = pd.concat([mol.steps, result], axis=1)
    return mol.steps

def analyze_steps(file, save=True):
    molecules_with_data = [mol for mol in file.molecules if mol.steps is not None]
    with concurrent.futures.ThreadPoolExecutor() as executor:
        results = [executor.submit(process_molecule, mol)
                   for mol in molecules_with_data]
    # if the results need to be manipulated uncomment the following lines:
    # for f in concurrent.futures.as_completed(results):
    #     print(f.result())

    filename = f'{file.relativeFilePath}_dwells_data.xlsx'
    data = file.savetoExcel(filename=filename, save=save)
    return data

if __name__ == '__main__':

    import sys
    from pathlib import Path
    p = Path(__file__).parents[2]
    sys.path.insert(0, str(p))

    from papylio.experiment import Experiment

    start = time.time()
    main_path='F:/Google Drive/PhD/Programming - Data Analysis/traceanalysis/traces'
    exp = Experiment(main_path)
    file = exp.files[0]

    data = analyze_steps(file)


    print(f'Processed step data in {time.time() - start:.2f} sec')





import numpy as np
import matplotlib.pyplot as plt
import matplotlib.path as pth
from pathlib import Path
import itertools
from scipy.spatial import cKDTree
import random
import matchpoint as mp
from papylio.plugins.sequencing.geometricHashing2 import polygon_area
import matchpoint as mp
from skimage.transform import AffineTransform
import time



class GeometricHashTable:
    def __init__(self, destinations=None, source_vertices=None, initial_source_transformation=AffineTransform(),
                 number_of_source_bases=20, number_of_destination_bases='all', load=False):
        # self.tile = tile
        # self.files = files # List of coordinate sets
        # self.dataPath = Path(dataPath)

        if load:
            self.load()
        else:

            self.initial_source_transformation = initial_source_transformation

            # self.mode = mode
            # if mode == 'translation': self.hashTableRange = [-10000, 10000]
            # else: self.hashTableRange = [-1,1]
            # self.nBins = nBins

            #
            # self.number_of_source_bases = number_of_source_bases
            # self.number_of_destination_bases = number_of_destination_bases

            # self._hashTable = None
            # self._matches = None
            #
            # self.rotationRange = rotationRange
            # self.magnificationRange = magnificationRange
            #


            self.destinations = destinations
            self.destination_KDTrees = [cKDTree(destination) for destination in destinations]
            self.source_vertices = source_vertices

            self.number_of_hashtable_entries_per_destination = []
            self.number_of_hashtable_entries_per_basis = []
            self.number_of_bases_per_destination = []

            self.basis_index_edges_for_destination = None
            self.entry_index_edges_for_destination = None
            self.entry_index_edges_for_basis = None

            self.hashtable = None

            self.create_hashtable()

    def save(self):
        # destinations_stacked = np.vstack([np.c_[i*np.ones(destination.shape[0], dtype=int), destination]
        #                                   for i, destination in enumerate(self.destinations)])
        destinations_stacked = np.vstack(self.destinations)
        destinations_length = np.array([len(destination) for destination in self.destinations])

        # number_of_hashtable_entries_per_basis_stacked = \
        #     np.vstack([np.c_[i * np.ones(len(n), dtype=int), n]
        #                for i, n in enumerate(self.number_of_hashtable_entries_per_basis)])

        number_of_hashtable_entries_per_basis_stacked = np.hstack(self.number_of_hashtable_entries_per_basis)
        number_of_hashtable_entries_per_basis_length = np.array([len(n) for n in self.number_of_hashtable_entries_per_basis])

        np.savez('geometric_hashtable.npz',
                    initial_source_transformation=self.initial_source_transformation.params,
                    destinations_stacked=destinations_stacked,
                    destinations_length=destinations_length,
                    hashtable=self.hashtable.data,
                    source_vertices=self.source_vertices,
                    number_of_hashtable_entries_per_destination=np.array(self.number_of_hashtable_entries_per_destination),
                    number_of_hashtable_entries_per_basis_stacked=number_of_hashtable_entries_per_basis_stacked,
                    number_of_hashtable_entries_per_basis_length=number_of_hashtable_entries_per_basis_length,
                    number_of_bases_per_destination=np.array(self.number_of_bases_per_destination),
                    basis_index_edges_for_destination=self.basis_index_edges_for_destination,
                    entry_index_edges_for_destination=self.entry_index_edges_for_destination,
                    entry_index_edges_for_basis=self.entry_index_edges_for_basis
                    )

    def load(self):
        file = np.load('geometric_hashtable.npz')
        self.initial_source_transformation = AffineTransform(file['initial_source_transformation'])
        self.destinations = np.split(file['destinations_stacked'], np.cumsum(file['destinations_length'])[:-1])
        self.destination_KDTrees = [cKDTree(destination) for destination in self.destinations]

        self.hashtable = cKDTree(file['hashtable'])
        self.source_vertices = file['source_vertices']

        self.number_of_hashtable_entries_per_destination = file['number_of_hashtable_entries_per_destination'].tolist()

        number_of_hashtable_entries_per_basis=\
            np.split(file['number_of_hashtable_entries_per_basis_stacked'],
                 np.cumsum(file['number_of_hashtable_entries_per_basis_length'])[:-1])
        self.number_of_hashtable_entries_per_basis = [n.tolist() for n in number_of_hashtable_entries_per_basis]

        self.number_of_bases_per_destination = file['number_of_bases_per_destination'].tolist()

        self.basis_index_edges_for_destination = file['basis_index_edges_for_destination']
        self.entry_index_edges_for_destination = file['entry_index_edges_for_destination']
        self.entry_index_edges_for_basis = file['entry_index_edges_for_basis']

    def create_hashtable(self):
        hashtable_entries_per_destination = []
        for destination in self.destinations:
            destination_bases = destination

            hashtable_entries_per_basis = [destination - basis for basis in destination_bases]

            if self.source_vertices is not None:
                # Make the area for which the hashtable is constructed twice the area of the source.
                # The area is still centered on the original source area
                center = np.mean(self.source_vertices, axis=0)
                crop_vertices_in_source = (self.source_vertices - center) * 2 + center
                crop_vertices_in_destination = self.initial_source_transformation(crop_vertices_in_source)
                hashtable_entries_per_basis = [mp.point_set.crop_coordinates(entries, crop_vertices_in_destination) for entries in hashtable_entries_per_basis]

            self.number_of_hashtable_entries_per_basis.append([len(entries) for entries in hashtable_entries_per_basis])

            self.number_of_bases_per_destination.append(len(destination))
            hashtable_entries_per_destination.append(np.vstack(hashtable_entries_per_basis))

        self.number_of_hashtable_entries_per_destination = [np.sum(c) for c in self.number_of_hashtable_entries_per_basis]

        self.basis_index_edges_for_destination = np.cumsum(np.hstack([[0], self.number_of_bases_per_destination]))
        self.entry_index_edges_for_destination = np.cumsum(np.hstack([[0], self.number_of_hashtable_entries_per_destination]))
        self.entry_index_edges_for_basis = np.cumsum(np.hstack([[0], np.hstack(self.number_of_hashtable_entries_per_basis)]))

        # self.destination_bases_index_bin_edges = np.cumsum([0]+[len(c) for i, c in enumerate(hash_table_entries_per_basis)])
        # or self.destination_bases_start_indices
        self.hashtable = cKDTree(np.vstack(hashtable_entries_per_destination))




    # def geometric_hash(point_sets, maximum_distance=100, tuple_size=4):
    #     # TODO: Add minimum_distance and implement
    #     # TODO: Make invariant to mirroring
    #     # TODO: Make usable with multiple point-sets in a single hash table
    #     # TODO: Implement names of point_sets, possibly through a dictionary and adding a attribute to each KDtree
    #
    #     if type(point_sets) is not list:
    #         point_sets = [point_sets]
    #
    #     point_set_KDTrees = [cKDTree(point_set) for point_set in point_sets]
    #
    #     hash_table = []
    #     point_tuple_sets = []
    #     for point_set_KDTree in point_set_KDTrees:
    #         point_tuple_generator = generate_point_tuples(point_set_KDTree, maximum_distance, tuple_size)
    #
    #         point_tuples = []
    #         for point_tuple, hash_code in geometric_hash_table(point_set_KDTree, point_tuple_generator, tuple_size):
    #             point_tuples.append(point_tuple)
    #             hash_table.append(hash_code)
    #
    #         point_tuple_sets.append(point_tuples)
    #
    #     hash_table_KDTree = cKDTree(np.array(hash_table))
    #
    #     return point_set_KDTrees, point_tuple_sets, hash_table_KDTree


    def query(self, source, distance=15, alpha=0.9, sigma=10, K_threshold=10e2):
        best_destination_basis_for_source_basis = []
        source_bases = source
        for basis in source_bases:
            hash_table_queries_in_source = source - basis
            hash_table_queries_in_destination = self.initial_source_transformation(hash_table_queries_in_source)
            found_destination_indices_per_query = self.hashtable.query_ball_point(hash_table_queries_in_destination, distance)
            # hash_table_queries_in_destination = cKDTree(hash_table_queries_in_destination)
            # hash_table_queries_in_destination = self.hashtable.query_ball_tree(hash_table_queries_in_destination, 15)
            found_destination_indices = np.hstack(found_destination_indices_per_query)

            count_per_destination_basis = np.histogram(found_destination_indices, bins=self.entry_index_edges_for_basis)[0]
            best_matching_destination_basis_index = np.argmax(count_per_destination_basis)
            max_count = np.max(count_per_destination_basis)
            best_destination_basis_for_source_basis.append((best_matching_destination_basis_index, max_count))

        # base_transformed = self.initial_image_transformation(np.array([[0,0]]))

        best_destination_basis_for_source_basis = \
            np.hstack([np.atleast_2d(np.arange(len(best_destination_basis_for_source_basis))).T,
                       best_destination_basis_for_source_basis])

        best_destination_basis_for_source_basis = best_destination_basis_for_source_basis[np.flip(np.argsort(best_destination_basis_for_source_basis[:, 2]))]

        for source_basis_index, destination_basis_index, count in best_destination_basis_for_source_basis:
        # i=-1
        # best_matching_source_basis_index = best_destination_basis_for_source_basis[i,0]
        # best_matching_destination_basis_index = best_destination_basis_for_source_basis[i,1]
            destination_index = np.where(destination_basis_index<self.basis_index_edges_for_destination[1:])[0][0]

            source_basis_in_destination = self.initial_source_transformation(source[source_basis_index])
            translation = self.destinations[destination_index][[destination_basis_index-self.basis_index_edges_for_destination[destination_index]]]\
                          - source_basis_in_destination

            found_transformation = self.initial_source_transformation + AffineTransform(translation=translation)

            # plt.figure()
            # mp.icp.scatter_coordinates([self.destination, found_transformation(source)])

            if self.test_transformation(source, self.destination_KDTrees[destination_index], found_transformation,
                                        alpha, sigma, K_threshold):
                match = mp.MatchPoint(source=source, destination=self.destinations[destination_index], method='Geometric hashing',
                                 transformation_type='linear', initial_transformation=None)
                match.transformation = found_transformation
                match.transformation_inverse = AffineTransform(matrix=found_transformation._inv_matrix)
                # match.calculate_inverse_transformation()
                match.destination_index = destination_index
                match.initial_transformation = self.initial_source_transformation
                match.source_vertices = self.source_vertices

                # match.hash_table_distance = distance
                #match.hash_table_distances_checked = hash_table_distances_checked
                #match.tuples_checked = tuples_checked
                return match

        else:
            return None


    def test_transformation(self, source, destination, found_transformation, alpha=0.9, sigma=10, K_threshold=10e2):
        if not type(destination) is cKDTree:
            destination_KDTree = cKDTree(destination)
        else:
            destination_KDTree = destination

        source_vertices_transformed = found_transformation(self.source_vertices)
        destination_cropped = mp.point_set.crop_coordinates(destination_KDTree.data, source_vertices_transformed)

        source_transformed_area = polygon_area(source_vertices_transformed)

        pDB = 1 / source_transformed_area

        K=1

        # TODO: Remove basis points?
        for coordinate in found_transformation(source):
            distance, index = destination_KDTree.query(coordinate)

            # 2d Gaussian
            pDF = alpha / source_transformed_area + \
                  (1 - alpha) / (2 * np.pi * sigma ** 2) * \
                  np.exp(-(distance ** 2) / (2 * sigma ** 2)) / len(destination_cropped)

            K = K * pDF/pDB
            if K > K_threshold:
                print("Found match")
                return True

        return False


if __name__ == '__main__':
    # # source = np.loadtxt(r'D:\SURFdrive\Promotie\Code\Python\papylio\papylio\plugins\sequencing\source.txt')
    # source = np.loadtxt(r'D:\SURFdrive\Promotie\Code\Python\papylio\papylio\plugins\sequencing\source3.txt')
    # destination1 = np.loadtxt(r'D:\SURFdrive\Promotie\Code\Python\papylio\papylio\plugins\sequencing\destination.txt')
    # destination2 = np.loadtxt(r'D:\SURFdrive\Promotie\Code\Python\papylio\papylio\plugins\sequencing\destination2.txt')
    # destinations = [destination1, destination2]
    #
    # initial_magnification = np.array([ 3.67058194, -3.67058194])
    # initial_rotation = 0.6285672733195177 # degrees
    #
    # initial_source_transformation = AffineTransform(matrix=None, scale=initial_magnification,
    #                                                 rotation=initial_rotation/360*np.pi*2,
    #                                                 shear=None, translation=None)
    # source_vertices = np.array([[256,   0], [512,   0], [512, 512], [256, 512]])
    #
    # ht = GeometricHashTable(destinations, source_vertices, initial_source_transformation=initial_source_transformation)
    #
    # match = ht.query(source, 15)

    from matchpoint.point_set_simulation import simulate_mapping_test_point_set

    # Simulate source and destination point sets
    number_of_source_points = 4000
    transformation = AffineTransform(translation=[128, 128], rotation=0 / 360 * 2 * np.pi, scale=[1, 1])
    source_bounds = np.array([[0, 0], [512, 512]])
    source_crop_bounds = np.array([[0, 0], [50, 50]])
    fraction_missing_source = 0.8
    fraction_missing_destination = 0.6
    maximum_error_source = 0
    maximum_error_destination = 0
    shuffle = True

    destination, source = simulate_mapping_test_point_set(number_of_source_points, transformation,
                                                          source_bounds, source_crop_bounds,
                                                          fraction_missing_source, fraction_missing_destination,
                                                          maximum_error_source, maximum_error_destination, shuffle)
    destinations = [destination]

    perfect = mp.MatchPoint(source, destination)
    perfect.transformation = AffineTransform(matrix=transformation._inv_matrix)
    perfect.show_mapping_transformation()

    # mp.icp.scatter_coordinates([source, destination])

    source_vertices = np.array([source_crop_bounds[0], source_crop_bounds.T[0],
                                source_crop_bounds[1], np.flip(source_crop_bounds.T[1])])
    ht = GeometricHashTable(destinations, source_vertices)

    distance = 5
    alpha = 0.5
    sigma = 2
    K_threshold = 10e8

    test = ht.query(source, distance, alpha, sigma, K_threshold)
    if test:
        test.show_mapping_transformation()



import itertools
import numpy as np

def all_sequence_subsets():
    bases = ['A', 'T', 'C', 'G']
    return np.array([''.join(bpc) for bpc in itertools.product(*[bases]*8)])

def all_basepaired_subsets():
    basepairs = ['AT', 'TA', 'CG', 'GC']
    return np.array([''.join(bpc)[1:] + ''.join(bpc)[0] for bpc in itertools.product(*[basepairs]*4)])

def basepaired_subsets():
    basepairs = ['AT', 'TA', 'CG', 'GC']
    basepaired_subsets = []
    for i0, bp0 in enumerate(basepairs):
        for i1, bp1 in enumerate(basepairs):
            for i2, bp2 in enumerate(basepairs):
                for i3, bp3 in enumerate(basepairs):
                    bpc = bp0 + bp1 + bp2 + bp3
                    basepaired_subsets.append(''.join(bpc)[1:] + ''.join(bpc)[0])
    return basepaired_subsets




def rotationally_symmetric_subsets(sequence_subset):
    return list(set([sequence_subset[i*2:] + sequence_subset[:i*2] for i in range(4)]))

def rotationally_symmetric_subset_categories(sequence_subsets):
    sequence_subsets = list(sequence_subsets)
    index = 0
    subset_category_dict = {}
    while len(sequence_subsets) > 0:
        sss = rotationally_symmetric_subsets(sequence_subsets[0])
        for ss in sss:
            subset_category_dict[ss] = index
            if ss in sequence_subsets:
                sequence_subsets.remove(ss)
        index += 1
    return subset_category_dict

def rotationally_symmetric_subset_groups():
    sequence_subsets = all_basepaired_subsets()
    sequence_subsets = list(sequence_subsets)
    index = 0
    subset_groups = []
    while len(sequence_subsets) > 0:
        sss = rotationally_symmetric_subsets(sequence_subsets[0])
        subset_groups.append(sss)
        for ss in sss:
            sequence_subsets.remove(ss)
    return subset_groups


def sequence_subset_with_comparable_structure(sequence_subset):
    # sequence_subset = sequence_subset[7] + sequence_subset[:7]
    sequence_subsets = []
    for i in range(4):
        sequence_subsets.append(sequence_subset[i*2:] + sequence_subset[:i*2])
    for ss in sequence_subsets.copy():
        change_base_within_basepair = str.maketrans('CG', 'GC')
        sequence_subsets.append(ss.translate(change_base_within_basepair))
    for ss in sequence_subsets.copy():
        change_base_within_basepair = str.maketrans('AT', 'TA')
        sequence_subsets.append(ss.translate(change_base_within_basepair))
    for ss in sequence_subsets.copy():
        switch_base_pairs = str.maketrans('ATCG', 'GCAT')
        sequence_subsets.append(ss.translate(switch_base_pairs))
    return set(sequence_subsets)


def sequence_subset_structure_category():
    subset_category_dict = {}
    sequence_subsets = basepaired_subsets()

    structure_index = 0
    while len(sequence_subsets) > 0:
        ss_with_structure = sequence_subset_with_comparable_structure(sequence_subsets[0])
        for ss in ss_with_structure:
            subset_category_dict[ss] = structure_index
            sequence_subsets.remove(ss)
        structure_index += 1
    return subset_category_dict


def unique_subset_structures():
    unique_subset_category_dict = {}
    for key, value in sequence_subset_structure_category().items():
        if value not in unique_subset_category_dict.values():
            unique_subset_category_dict[key] = value
    return unique_subset_category_dict



import os
import numpy as np

from papylio.movie.movie import Movie


class NskMovie(Movie):
    extensions = ['.nsk']

    def __init__(self, arg, *args, **kwargs):
        super().__init__(arg, *args, **kwargs)

        self.writepath = self.filepath.parent
        self.name = self.filepath.with_suffix('').name
        self.rot90 = 1

        self.channel_arrangement = np.array([[[0, 1]]])

        self.data_type = np.dtype(np.uint16)

        # self.read_header()
        self.create_frame_info()  # Possibly move to Movie later on

        # self._initialized = True

    def _read_header(self):
        with self.filepath.open('rb') as fid:
            self.width = int(np.fromfile(fid, dtype=np.int16, count=1)[0])
            self.height = int(np.fromfile(fid, dtype=np.int16, count=1)[0])

        self.number_of_frames = int((os.path.getsize(self.filepath) - 4) / 2 / self.width / self.height)

    def _read_frame(self, frame_number):
        with self.filepath.open('rb') as fid:
            fid.seek(4 + 2 * frame_number * int(self.width * self.height), os.SEEK_SET)
            image = np.fromfile(fid, dtype=np.uint16, count=self.width*self.height)
            image = np.reshape(image, (self.width, self.height))

        return image

    def _read_frames(self, indices):
        # Can probably be implemented more efficiently
        return np.stack([self._read_frame(i) for i in indices])


import pytest
import xarray as xr
import tifffile
import numpy as np
from PySide2.QtWidgets import QApplication
import sys
from multiprocessing import Process, freeze_support

from pathlib2 import Path

# print(__file__)

# sys.path.append(Path(__file__).parent.parent.parent)
from papylio.gui.main import MainWindow

def test_GUI(shared_datadir):
    print(shared_datadir)
    freeze_support()

    app = QApplication(sys.argv)

    window = MainWindow(shared_datadir / 'BN_TIRF')
    window.show()

    app.exec_()





import sys
import PySide2

import sys
from PySide2.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QGridLayout, QTreeView, QApplication, QMainWindow, \
    QPushButton, QTabWidget, QTableWidget, QComboBox, QLineEdit
from PySide2.QtGui import QStandardItem, QStandardItemModel
from PySide2.QtCore import Qt

import matplotlib as mpl
from matplotlib.backends.backend_qtagg import (
    FigureCanvas, NavigationToolbar2QT as NavigationToolbar)
import matplotlib.pyplot as plt
import numpy as np

from papylio import Experiment, File
from papylio.trace_plot import TracePlotWindow
from papylio.gui.selection_widget import SelectionWidget

# class TreeNode:
#     def __init__(self, node_object, parent=None):
#         self.parent = parent
#         if isinstance(node_object, Experiment):
#             self.experiment = node_object
#             self.name = self.experiment.name
#             self.type = 'experiment'
#         elif isinstance(node_object, str):
#             self.name = node_object
#             self.type = 'folder'
#         elif isinstance(node_object, File):
#             self.file = node_object
#             self.name = self.file.name
#             self.type = 'file'
#
#         self.children = []
#
#     def data(self, column):
#         # if column == 0:
#         return self.columnValues[column]
#         # else:
#         #     return ''
#         # return self._data[column]
#
#     def appendChild(self, node_object):
#         node = TreeNode(node_object, self)
#         self.children.append(node)
#         return node
#
#     def child(self, row):
#         return self.children[row]
#
#     def childrenCount(self):
#         return len(self.children)
#
#     def hasChildren(self):
#         if len(self.children) > 0:
#             return True
#         return False
#
#     def row(self):
#         if self.parent is not None:
#             return self.parent.children.index(self)
#         else:
#             return 0
#
#     @property
#     def columnValues(self):
#         return [self.name]
#
#     def columnCount(self):
#         return len(self.columnValues)
#
#     def __repr__(self):
#         return f'TreeNode: {self.name}'
#
#
# class TreeModel(QAbstractItemModel):
#     def __init__(self, parent=None):
#         super().__init__(parent)
#         # column_names = ['Column1','Column2']
#         self.root = TreeNode('Name')
#         self.createData()
#         print('t')
#
#     def createData(self):
#         for x in ['a','b','c']:
#             self.root.appendChild(x)
#         for y in ['q','r','s']:
#             self.root.child(0).appendChild(y)
#         for z in ['d','e','f']:
#             self.root.child(2).appendChild(z)
#
#     def addExperiment(self, experiment):
#         # experiment = Experiment(r'D:\SURFdrive\Promotie\Code\Python\papylio\twoColourExampleData\20141017 - Holliday junction - Copy')
#         #experiment = Experiment(r'C:\Users\ivoseverins\surfdrive\Promotie\Code\Python\papylio\twoColourExampleData\20141017 - Holliday junction - Copy')
#         experimentNode = self.root.appendChild(experiment)
#         for file in experiment.files:
#             print('addfile'+file.name)
#             self.addFile(file, experimentNode)
#
#         print('add')
#
#     def addFile(self, file, experimentNode):
#         # pass
#
#         folders = file.relativePath.parts
#
#         #nodeItemNames = [item.GetText() for item in experimentNode.children if item.GetData() == None]
#
#         parentItem = experimentNode
#         for folder in folders:
#
#             # Get the folderItems and folder names for the current folderItem
#             nodeItems = [item for item in parentItem.children if item.type == 'folder']
#             nodeItemNames = [item.name for item in nodeItems]
#
#             if folder not in nodeItemNames:
#                 # Add new item for the folder and set parentItem to this item
#                 parentItem = parentItem.appendChild(folder)
#             else:
#                 # Set parent item to the found folderItem
#                 parentItem = nodeItems[nodeItemNames.index(folder)]
#
#         item = parentItem.appendChild(file)
#         #self.FileItems.append(item)
#
#         # self.insertDataIntoColumns(item)
#
#         return item
#
#     def columnCount(self, index=QtCore.QModelIndex()):
#         if index.isValid():
#             return index.internalPointer().columnCount()
#         else:
#             return self.root.columnCount()
#
#     def rowCount(self, index=QtCore.QModelIndex()):
#         if index.row() > 0:
#             return 0
#         if index.isValid():
#             item = index.internalPointer()
#         else:
#             item = self.root
#         return item.childrenCount()
#
#     def index(self, row, column, index=QtCore.QModelIndex()):
#         if not self.hasIndex(row, column, index):
#             return QtCore.QModelIndex()
#         if not index.isValid():
#             item = self.root
#         else:
#             item = index.internalPointer()
#
#         child = item.child(row)
#         if child:
#             return self.createIndex(row, column, child)
#         return QtCore.QMOdelIndex()
#
#     def parent(self, index):
#         if not index.isValid():
#             return QtCore.QModelIndex()
#         item = index.internalPointer()
#         if not item:
#             return QtCore.QModelIndex()
#
#         parent = item.parent
#         if parent == self.root:
#             return QtCore.QModelIndex()
#         else:
#             return self.createIndex(parent.row(), 0, parent)
#
#     def hasChildren(self, index):
#         if not index.isValid():
#             item = self.root
#         else:
#             item = index.internalPointer()
#         return item.hasChildren()
#
#     def data(self, index, role=QtCore.Qt.DisplayRole):
#        if index.isValid() and role == QtCore.Qt.DisplayRole:
#             return index.internalPointer().data(index.column())
#        elif not index.isValid():
#             return self.root.getData()
#
#     def headerData(self, section, orientation, role):
#         if orientation == QtCore.Qt.Horizontal and role == QtCore.Qt.DisplayRole:
#             return self.root.data(section)
#
#
#
# class MainWindow(QMainWindow):
#     def __init__(self):
#         super().__init__()
#         # model = QFileSystemModel()
#         # model.setRootPath(QDir.currentPath())
#
#
#
#         self.model = TreeModel()
#
#         self.tree = QTreeView()
#         self.tree.setModel(self.model)
#
#         from papylio import Experiment
#         experiment = Experiment(r'D:\SURFdrive\Promotie\Code\Python\papylio\twoColourExampleData\20141017 - Holliday junction - Copy')
#         #experiment = Experiment(r'C:\Users\ivoseverins\surfdrive\Promotie\Code\Python\papylio\twoColourExampleData\20141017 - Holliday junction - Copy')
#         #self.model.addExperiment(experiment)
#
#         self.setCentralWidget(self.tree)


class MainWindow(QMainWindow):
    # def __init__(self):
    #     super().__init__()
    #     # model = QFileSystemModel()
    #     # model.setRootPath(QDir.currentPath())
    #
    #
    #
    #     self.model = TreeModel()
    #
    #     self.tree = QTreeView()
    #     self.tree.setModel(self.model)
    #
    #      #experiment = Experiment(r'C:\Users\ivoseverins\surfdrive\Promotie\Code\Python\papylio\twoColourExampleData\20141017 - Holliday junction - Copy')
    #     #self.model.addExperiment(experiment)
    #
    #     self.setCentralWidget(self.tree)

    def __init__(self, main_path=None):
        super().__init__()

        from papylio import Experiment
        # self.experiment = Experiment(
        #     r'D:\SURFdrive\Promotie\Code\Python\papylio\twoColourExampleData\20141017 - Holliday junction - Copy')
        self.experiment = Experiment(main_path)

        self.tree = QTreeView(self)
        layout = QVBoxLayout()
        layout.addWidget(self.tree)
        self.model = QStandardItemModel()
        self.root = self.model.invisibleRootItem()
        self.model.setHorizontalHeaderLabels(['Name', 'Count'])
        self.tree.header().setDefaultSectionSize(180)
        self.tree.setModel(self.model)
        self.addExperiment(self.experiment)
        self.tree.setFocusPolicy(Qt.NoFocus)
        self.tree.setFixedWidth(256)
        self.update = True

        self.model.itemChanged.connect(self.onItemChange)


        self.image_canvas = ImageCanvas(self, width=5, height=4, dpi=100)

        # Create toolbar, passing canvas as first parament, parent (self, the MainWindow) as second.
        image_toolbar = NavigationToolbar(self.image_canvas, self)

        image_layout = QVBoxLayout()
        image_layout.addWidget(image_toolbar)
        image_layout.addWidget(self.image_canvas)

        # Create a placeholder widget to hold our toolbar and canvas.
        self.image = QWidget()
        self.image.setLayout(image_layout)

        controls_layout = QGridLayout()
        controls_layout.setAlignment(Qt.AlignTop)

        # controls_layout.addWidget(QLabel('Minimum intensity difference'), 0, 0)
        # mid = QLineEdit(str(self.experiment.configuration['find_coordinates']['peak_finding']['minimum_intensity_difference']))
        # mid.textChanged.connect(self.midChange)
        # controls_layout.addWidget(mid, 0, 1)

        perform_mapping_button = QPushButton('Perform mapping')
        perform_mapping_button.clicked.connect(self.perform_mapping)
        controls_layout.addWidget(perform_mapping_button, 1, 0, 1, 2)

        find_molecules_button = QPushButton('Find coordinates')
        find_molecules_button.clicked.connect(self.find_coordinates)
        controls_layout.addWidget(find_molecules_button, 2, 0, 1, 2)

        extract_traces_button = QPushButton('Extract traces')
        extract_traces_button.clicked.connect(self.extract_traces)
        controls_layout.addWidget(extract_traces_button, 3, 0, 1, 2)

        self.controls = QWidget()
        self.controls.setLayout(controls_layout)
        self.controls.setMinimumWidth(200)


        extraction_layout = QHBoxLayout()
        extraction_layout.addWidget(self.image)
        extraction_layout.addWidget(self.controls)




        # self.selection = QTableWidget()
        # self.selection.setRowCount(5)
        # self.selection.setColumnCount(4)


        tabs = QTabWidget()
        tabs.setTabPosition(QTabWidget.North)
        tabs.setMovable(False)
        tabs.setDocumentMode(True)

        tab1 = QWidget(self)
        tab1.setLayout(extraction_layout)
        tabs.addTab(tab1, 'Movie')
        self.traces = TracePlotWindow(parent=self, width=4, height=3, show=False,
                                      save_path=self.experiment.analysis_path.joinpath('Trace_plots'))
        tabs.addTab(self.traces, 'Traces')
        self.selection = SelectionWidget()
        tabs.addTab(self.selection, 'Selection (beta)')
        tabs.currentChanged.connect(self.setTabFocus)

        experiment_layout = QVBoxLayout()

        refresh_button = QPushButton('Refresh')
        refresh_button.clicked.connect(self.refresh)
        experiment_layout.addWidget(refresh_button)

        experiment_layout.addWidget(self.tree)

        layout = QHBoxLayout()
        layout.addLayout(experiment_layout)
        layout.addWidget(tabs)

        widget = QWidget()
        widget.setLayout(layout)
        self.setCentralWidget(widget)

    def keyPressEvent(self, e):
        self.traces.keyPressEvent(e)

    def setTabFocus(self, e):
        if e == 0:
            self.image.setFocus()
        if e == 1:
            self.traces.setFocus()

    def midChange(self, input):
        input = int(input)
        self.experiment.configuration['find_coordinates']['peak_finding']['minimum_intensity_difference'] = input
        self.experiment.configuration.save()

    def perform_mapping(self, t):
        print(t)
        selected_files = self.experiment.selectedFiles
        if selected_files:
            selected_files.serial.perform_mapping()
            self.image_canvas.refresh()
            plt.show()

    def find_coordinates(self):
        selected_files = self.experiment.selectedFiles
        if selected_files:
            selected_files.movie.determine_spatial_background_correction(use_existing=True)
            selected_files.find_coordinates()
            self.image_canvas.refresh()
            self.update_plots()

    def extract_traces(self):
        selected_files = self.experiment.selectedFiles
        if selected_files:
            selected_files.extract_traces()
            # self.image_canvas.refresh()
            self.update_plots()

    def onItemChange(self, item):
        if isinstance(item.data(), File):
            file = item.data()
            file.isSelected = (True if item.checkState() == Qt.Checked else False)
            print(f'{file}: {file.isSelected}')

        else:
            self.update = False
            for i in range(item.rowCount()):
                item.child(i).setCheckState(item.checkState())
            self.update = True

        if self.update:
            self.update_plots()

    def update_plots(self):
        selected_files = self.experiment.selectedFiles + [None]
        self.image_canvas.file = selected_files[0]
        if selected_files[0] is not None:
            self.traces.dataset = selected_files[0].dataset
            self.selection.file = selected_files[0]
        else:
            self.traces.dataset = None
            self.selection.file = None

    def addExperiment(self, experiment):

        # experiment = Experiment(r'D:\SURFdrive\Promotie\Code\Python\papylio\twoColourExampleData\20141017 - Holliday junction - Copy')
        #experiment = Experiment(r'C:\Users\ivoseverins\surfdrive\Promotie\Code\Python\papylio\twoColourExampleData\20141017 - Holliday junction - Copy')
        self.root.appendRow([
                QStandardItem(experiment.name),
                QStandardItem(0),
            ])
        experimentNode = self.root.child(self.root.rowCount() - 1)
        for file in experiment.files:
            print('addfile'+file.name)
            self.addFile(file, experimentNode)

        self.tree.expandAll()

        print('add')

    def addFile(self, file, experimentNode):
        folders = file.relativePath.parts

        parentItem = experimentNode
        parentItem.setCheckable(True)
        for folder in folders:

            # Get the folderItems and folder names for the current folderItem
            nodeItems = [parentItem.child(i) for i in range(parentItem.rowCount())]# if item.type == 'folder']
            nodeItemNames = [item.text() for item in nodeItems]

            if folder not in nodeItemNames:
                # Add new item for the folder and set parentItem to this item
                parentItem.appendRow([
                    QStandardItem(folder),
                    QStandardItem(0),
                ])
                parentItem = parentItem.child(parentItem.rowCount() - 1)
                parentItem.setCheckable(True)
            else:
                # Set parent item to the found folderItem
                parentItem = nodeItems[nodeItemNames.index(folder)]

        parentItem.appendRow([
            QStandardItem(file.name),
            QStandardItem(0),
        ])
        item = parentItem.child(parentItem.rowCount() - 1)
        item.setCheckable(True)
        if file.isSelected:
            item.setCheckState(Qt.Checked)
        else:
            item.setCheckState(Qt.Unchecked)
        item.setData(file)
        #self.FileItems.append(item)

        # self.insertDataIntoColumns(item)

        return item

    def refresh(self):
        self.root.removeRows(0, 1)
        self.experiment = Experiment(self.experiment.main_path)
        self.addExperiment(self.experiment)


class ImageCanvas(FigureCanvas):
    def __init__(self, parent=None, width=14, height=7, dpi=100):
        self.figure = mpl.figure.Figure(figsize=(width, height), dpi=dpi, constrained_layout=True)  # , figsize=(2, 2))
        super().__init__(self.figure)
        self.parent = parent

        # self.axis = self.figure.gca()

        self._file = None

    @property
    def file(self):
        return self._file

    @file.setter
    def file(self, file):
        if file is not None and file is not self._file:
            self._file = file
            self.refresh()
        elif file is None:
            self._file = None
            self.figure.clf()
            self.draw()

    def refresh(self):
        self.figure.clf()
        self._file.movie.determine_spatial_background_correction(use_existing=True)
        self._file.show_coordinates_in_image(figure=self.figure)
        self.draw()


if __name__ == '__main__':
    from multiprocessing import Process, freeze_support
    freeze_support()

    app = QApplication(sys.argv)

    window = MainWindow()
    window.show()

    app.exec_()



import re
import numpy as np
import xarray as xr


def datatype_conversion(dtype=list):
    def datatype_conversion_decorator(func):
        def wrapper(sequences, *args, **kwargs):
            if isinstance(sequences, xr.DataArray):
                coords = sequences.coords
                sequences = dtype(sequences.values)
                return xr.DataArray(func(sequences, *args, **kwargs), coords=coords)
            elif isinstance(sequences, np.ndarray):
                sequences = dtype(sequences)
                return np.array(func(sequences, *args, **kwargs))
            else:
                sequences = dtype(sequences)
                return list(func(sequences, *args, **kwargs))
        return wrapper
    return datatype_conversion_decorator


@datatype_conversion()
def fraction_GC(sequences):
   return [len(re.findall('G|C', sequence))/len(sequence) for sequence in sequences]

@datatype_conversion()
def number_of_neighboring_bases(sequences, base_type):
    if base_type == 'purine':
        search_string = 'AG|GA|AA|GG'
    elif base_type == 'pyrimidine':
        search_string = 'TC|CT|TT|CC'
    elif len(base_type) == 1:
        search_string = base_type + base_type
    return [len(re.findall(search_string, sequence + sequence[0])) for sequence in sequences]

@datatype_conversion()
def number_of_bases(sequences, base_type, positions='all'):
    if not positions=='all':
        sequences_new = []
        sequences_new.append([])
    if base_type == 'purine':
        search_string = 'AG|GA|AA|GG'
    elif base_type == 'pyrimidine':
        search_string = 'TC|CT|TT|CC'
    elif len(base_type) == 1:
        search_string = base_type + base_type
    return [len(re.findall(search_string, sequence + sequence[0])) for sequence in sequences]

def get_bases(bases):
    if bases == 'all':
        bases = 'ATCG' # np.array(['A', 'T', 'C', 'G'])
    elif bases == 'purines':
        bases = 'AG' #np.array(['A', 'G'])
    elif bases == 'pyrimidines':
        bases = 'TC' #np.array(['T', 'C'])

    # if return_type == str:
    #     ''.join(list(bases))
    # else:
    return np.array(list(bases))

import itertools
def get_base_combinations(base_combinations='basepaired'):
    if base_combinations in ['all','purines','pyrimidines']:
        base_combinations = np.array(list(itertools.product(get_bases(base_combinations),repeat=2)))
    elif base_combinations == 'basepaired':
        base_combinations = sequences_to_sequence_array(np.array(['AT', 'TA', 'GC', 'CG']))
    elif isinstance(base_combinations, str):
        base_combinations = sequences_to_sequence_array([base_combinations])
    else:
        base_combinations = sequences_to_sequence_array(base_combinations)
    return base_combinations

def sequences_to_sequence_array(sequences):
    return np.atleast_1d(sequences).astype('U').view('U1').reshape(-1, len(sequences[0]))

@datatype_conversion(np.array)
def base_count(sequences, positions='all', bases='all'):
    # TODO: make it possible to
    if positions == 'all':
        positions = np.arange(len(sequences[0]))
    else:
        positions = np.array(positions)

    sequence_array = sequences_to_sequence_array(sequences)

    return (sequence_array[:, positions, None] == bases[None, None, :]).any(axis=2).sum(axis=1)


def base_combination_presence(sequences, position_0, position_1, base_combinations):
    base_combinations = get_base_combinations(base_combinations)
    sequence_array = sequences_to_sequence_array(sequences)
    return (sequence_array[:, [position_0, position_1], None] == base_combinations.T[None, :, :]).all(axis=1).any(axis=1)

    # base_combinations_per_pair = np.stack(
    #     [get_base_combinations(base_combinations).T for base_combinations in base_combinations_per_pair])
@datatype_conversion(np.array)
def base_combination_count(sequences, position_pairs, base_combinations_per_pair):
    if isinstance(base_combinations_per_pair, str):
        base_combinations_per_pair = [base_combinations_per_pair] * len(position_pairs)
    return np.array([base_combination_presence(sequences, *position_pair, base_combinations) \
        for position_pair, base_combinations in zip(position_pairs, base_combinations_per_pair)]).sum(axis=0)

    # base_combinations_per_pair = np.stack([get_base_combinations(base_combinations).T for base_combinations in base_combinations_per_pair])
    # sequence_array = sequences_to_sequence_array(sequences)
    # return (sequence_array[:, position_pairs, None] == base_combinations_per_pair[None, :, :, :]).all(axis=2).sum(axis=2)
    #
    #
    #     base_combination_count = \
    #         np.array([(base_count(sequences, positions=position_pair, bases=bases,
    #                               count_basepairs=count_basepairs) == 2).astype(int)
    #                   for position_pair in position_pairs])
    #     return base_combination_count.sum(axis=0)


# @datatype_conversion(np.array)
# def base_combination_count(sequences, position_pairs, bases='purines'):
#     base_combination_count = \
#         np.array([(base_count(sequences, positions=position_pair, bases=bases, count_basepairs=count_basepairs)==2).astype(int)
#                   for position_pair in position_pairs])
#     return base_combination_count.sum(axis=0)
